from ..mpditem import MPDItem


class SegmentList(MPDItem):

    def __init__(self, name: str):
        super(SegmentList, self).__init__(name)
        self.timescale = 0 # type: int
        self.duration = 0 # type: int

    def generate(self):
        self.to_int()

    def to_int(self):
        try:
            self.timescale = int(self.timescale)
            self.duration = int(self.duration)
        except Exception:
            pass


from hlsdl.extractors.metaitem import MetaItem


class ISMItem(MetaItem):
    def __init__(self, name: str = "ISMItem"):
        self.name = name
        self.innertext = ''
        self.childs = []

    def addattr(self, name: str, value):
        self.__setattr__(name, value)

    def addattrs(self, attrs: dict):
        for attr_name, attr_value in attrs.items():
            attr_name: str
            attr_name = attr_name.replace(":", "_")
            self.addattr(attr_name, attr_value)

    def find(self, name: str):
        return [child for child in self.childs if child.name == name]

    def to_int(self, attr_name: str):
        value = self.__getattribute__(attr_name) # type: str
        if isinstance(value, str) and value.isdigit():
            self.__setattr__(attr_name, int(value))


from ..mpditem import MPDItem


class BaseURL(MPDItem):
    def __init__(self, name: str):
        super(BaseURL, self).__init__(name)
        self.serviceLocation = None # type: str
        self.dvb_priority = None # type: str
        self.dvb_weight = None # type: str


import re
from pathlib import Path
from typing import List, Union
from hlsdl.models.base import BaseUri
from hlsdl.models.stream import Stream
from hlsdl.util.maps.codecs import AUDIO_CODECS
from hlsdl.extractors.dash.segment import DASHSegment


class DASHStream(Stream):
    def __init__(self, index: int, uri_item: BaseUri, save_dir: Path):
        super(DASHStream, self).__init__(index, uri_item, save_dir)
        self.model = 'dash'
        self.segments = [] # type: List[DASHSegment]
        self.suffix = '.mp4'
        self.has_init_segment = False
        self.skey = '' # type: str
        self.append_segment()

    def get_name(self):
        if self.stream_type != '':
            base_name = f'{self.name}_{self.stream_type}'
        else:
            base_name = self.name
        if 'ixigua.com' in self.home_url:
            base_name += f'_{self.index}'
        if self.codecs is not None and self.codecs != '':
            base_name += f'_{self.codecs}'
        if self.stream_type == 'subtitle' and self.lang != '':
            base_name += f'_{self.lang}'
        if self.stream_type == 'subtitle' and self.skey != '':
            base_name += f'_{self.skey}'
        elif self.stream_type == 'text' and self.lang != '':
            base_name += f'_{self.lang}'
        elif self.stream_type == 'video' and self.resolution != '':
            base_name += f'_{self.resolution}'
        elif self.stream_type == 'audio' and self.lang != '':
            base_name += f'_{self.lang}'
        if self.stream_type in ['audio', 'video'] and self.bandwidth is not None:
            base_name += f'_{self.bandwidth / 1000:.2f}kbps'
        if self.stream_type == '' and self.skey:
            base_name += f'_{self.skey}'
        return base_name

    def append_segment(self):
        index = len(self.segments)
        if self.has_init_segment:
            index -= 1
        segment = DASHSegment().set_index(index).set_folder(self.save_dir)
        self.segments.append(segment)

    def update(self, stream: 'DASHStream', name_from_url: bool = False):
        '''
        Representation id相同可以合并
        这个时候应该重新计算时长和码率
        '''
        total_duration = self.duration + stream.duration
        if total_duration > 0:
            self.bandwidth = (stream.duration * stream.bandwidth + self.duration * self.bandwidth) / (self.duration + stream.duration)
        self.duration += stream.duration
        has_init = False
        for segment in stream.segments:
            # 被合并的流的init分段 避免索引计算错误
            if segment.segment_type == 'init':
                has_init = True
                stream.segments.remove(segment)
                break
        self.segments_extend(stream.segments, has_init=has_init, name_from_url=name_from_url)

    def set_subtitle_url(self, url: str):
        self.has_init_segment = True
        self.segments[-1].set_subtitle_url(self.fix_url(url))
        # self.append_segment()

    def set_init_url(self, url: str):
        self.has_init_segment = True
        self.segments[-1].set_init_url(self.fix_url(url))
        self.append_segment()

    def set_media_url(self, url: str, name_from_url: bool = False):
        self.segments[-1].set_media_url(self.fix_url(url), name_from_url=name_from_url)
        self.append_segment()

    def base2url(self, duration: float):
        if duration is not None:
            self.duration = duration
        self.segments[-1].set_media_url(self.base_url)
        self.append_segment()

    def set_segment_duration(self, duration: float):
        self.segments[-1].set_duration(duration)

    def set_segment_fmt_time(self, fmt_time: int):
        self.segments[-1].set_fmt_time(fmt_time)

    def set_segments_duration(self, duration: float):
        '''' init分段没有时长 这里只用设置普通分段的 '''
        for segment in self.segments:
            if segment.segment_type == 'init':
                continue
            segment.set_duration(duration)

    def get_skey(self):
        return self.skey

    def set_skey(self, aid: str, rid: str):
        _patch = ''
        if aid is not None:
            self.skey += f'{aid}'
            _patch = '_'
        if rid is not None:
            self.skey += f'{_patch}{rid}'.replace('/', '_')

    def set_lang(self, lang: str):
        if lang is None:
            return
        self.lang = lang

    def set_bandwidth(self, bandwidth: Union[str, int]):
        if bandwidth is None:
            return
        if isinstance(bandwidth, str):
            bandwidth = int(bandwidth)
        self.bandwidth = bandwidth

    def set_codecs(self, codecs: str):
        if codecs is None:
            return
        # https://chromium.googlesource.com/chromium/src/media/+/master/base/mime_util_internal.cc
        if re.match('avc(1|3)*', codecs):
            codecs = 'H264'
        if re.match('(hev|hvc)1*', codecs):
            codecs = 'H265'
        if re.match('vp(09|9)*', codecs):
            codecs = 'VP9'
        if codecs in ['wvtt', 'ttml', 'stpp']:
            codecs = codecs.upper()
        if AUDIO_CODECS.get(codecs) is not None:
            if 'AAC' in AUDIO_CODECS[codecs]:
                codecs = 'AAC'
            else:
                codecs = AUDIO_CODECS[codecs]
        self.codecs = codecs

    def set_resolution(self, width: str, height: str):
        if width is None or height is None:
            return
        self.resolution = f'{width}x{height}'
        self.skey += '_' + self.resolution

    def set_stream_type(self, stream_type: str):
        if stream_type is None:
            return
        if stream_type == 'subtitle':
            self.stream_type = stream_type
            return
        stream_type, stream_suffix = stream_type.split('/')
        if stream_suffix == 'ttml+xml':
            stream_type = 'subtitle'
            self.suffix = '.ttml'
        elif stream_suffix == 'vtt':
            self.suffix = '.vtt'
            stream_type = 'subtitle'
        if stream_type == 'application':
            if self.codecs is None:
                return
            if self.codecs.lower() in ['wvtt', 'ttml', 'stpp']:
                stream_type = 'subtitle'
            else:
                return
        self.stream_type = stream_type
        if self.stream_type == 'audio':
            self.suffix = '.m4a'


from ..mpditem import MPDItem


class Period(MPDItem):
    def __init__(self, name: str):
        super(Period, self).__init__(name)
        self.id = None # type: str
        self.start = None # type: float
        self.duration = None # type: float

    def generate(self):
        if isinstance(self.start, str):
            self.start = self.match_duration(self.start)
        # else:
        #     self.start = 0.0
        if isinstance(self.duration, str):
            self.duration = self.match_duration(self.duration)
        # else:
        #     self.duration = 0.0


from ..mpditem import MPDItem


class Representation(MPDItem):
    def __init__(self, name: str):
        super(Representation, self).__init__(name)
        self.id = None
        self.scanType = None
        self.frameRate = None
        self.bandwidth = None
        self.codecs = None
        self.mimeType = None
        self.sar = None
        self.width = None
        self.height = None
        self.audioSamplingRate = None

    def get_contenttype(self):
        if self.mimeType is not None:
            return self.mimeType.split('/')[0].title()

    def get_resolution(self):
        return f"{self.width}x{self.height}p"

    def get_suffix(self):
        return '.' + self.mimeType.split('/')[0].split('-')[-1]


from ..mpditem import MPDItem


class ContentProtection(MPDItem):
    def __init__(self, name: str):
        super(ContentProtection, self).__init__(name)
        self.value = None
        self.schemeIdUri = None
        self.cenc_default_KID = None


import os
import sys
import logging
import datetime
from pathlib import Path


GLOBAL_LOGGERS = {}


def tell_me_path(target_folder_name: str) -> Path:
    '''
    兼容自用脚本版与免安装可执行版本
    :param target_folder_name: 目录文件夹路径
    :returns 返回request_config路径
    '''
    if getattr(sys, 'frozen', False):
        return Path(sys.executable).parent / target_folder_name
    else:
        return Path(__file__).parent.parent / target_folder_name


class PackagePathFilter(logging.Filter):
    '''
    获取文件相对路径而不只是文件名
    配合行号可以快速定位
    参见 How can I include the relative path to a module in a Python logging statement?
    - https://stackoverflow.com/questions/52582458
    '''
    def filter(self, record):
        record.relativepath = None
        abs_sys_paths = map(os.path.abspath, sys.path)
        for path in sorted(abs_sys_paths, key=len, reverse=True):
            if not path.endswith(os.sep):
                path += os.sep
            if getattr(sys, 'frozen', False):
                record.relativepath = record.pathname
                break
            if record.pathname.startswith(path):
                record.relativepath = os.path.relpath(record.pathname, path)
                break
        return True


def setup_logger(name: str, level: str = 'INFO') -> logging.Logger:
    '''
    - 终端只输出日志等级大于等于level的日志 默认是INFO
    - 全部日志等级的信息都会记录到文件中
    '''
    logger = GLOBAL_LOGGERS.get(name)
    if logger:
        return logger
    # 先把 logger 初始化好
    logger = logging.getLogger(f'{name}')
    GLOBAL_LOGGERS[name] = logger
    # 开始设置
    formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(relativepath)s:%(lineno)d %(levelname)s: %(message)s', datefmt='%H:%M:%S')
    # formatter = logging.Formatter('%(asctime)s %(process)d %(relativepath)s:%(lineno)d %(levelname)s: %(message)s')
    log_time = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")
    # 没有打包的时候 __file__ 就是当前文件路径
    # 打包之后通过 sys.executable 获取程序路径
    log_folder_path = tell_me_path('logs')
    if log_folder_path.exists() is False:
        log_folder_path.mkdir()
    ch = logging.StreamHandler()
    ch.addFilter(PackagePathFilter())
    if level.lower() == 'info':
        ch.setLevel(logging.INFO)
    elif level.lower() == 'warning':
        ch.setLevel(logging.WARNING)
    elif level.lower() == 'error':
        ch.setLevel(logging.ERROR)
    else:
        ch.setLevel(logging.DEBUG)
    ch.setFormatter(formatter)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(ch)
    log_file_path = log_folder_path / f'{name}-{log_time}.log'
    fh = logging.FileHandler(log_file_path.resolve().as_posix(), encoding='utf-8', delay=True)
    fh.addFilter(PackagePathFilter())
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    return logger


def test_log():
    logger = setup_logger("test", level='INFO')
    logger.debug('this is DEBUG level')
    logger.info('this is INFO level')
    logger.warning('this is WARNING level')
    logger.error('this is ERROR level')


# test_log()


from xml.parsers.expat import ParserCreate
from .mpd import MPD
from .childs.location import Location
from .childs.adaptationset import AdaptationSet
from .childs.baseurl import BaseURL
from .childs.cencpssh import CencPssh
from .childs.contentprotection import ContentProtection
from .childs.period import Period
from .childs.representation import Representation
from .childs.role import Role
from .childs.s import S
from .childs.segmentlist import SegmentList
from .childs.initialization import Initialization
from .childs.segmenturl import SegmentURL
from .childs.segmentbase import SegmentBaee
from .childs.segmenttemplate import SegmentTemplate
from .childs.segmenttimeline import SegmentTimeline


def xml_handler(content: str):
    def handle_start_element(tag, attrs):
        nonlocal mpd
        nonlocal mpd_handlers
        if mpd is None:
            if tag != 'MPD':
                raise Exception('the first tag is not MPD!')
            mpd = MPD(tag)
            mpd.addattrs(attrs)
            mpd.generate()
            stack.append(mpd)
        else:
            if mpd_handlers.get(tag) is None:
                return
            child = mpd_handlers[tag](tag)
            child.addattrs(attrs)
            child.generate()
            mpd.childs.append(child)
            mpd = child
            stack.append(child)

    def handle_end_element(tag):
        nonlocal mpd
        nonlocal mpd_handlers
        if mpd_handlers.get(tag) is None:
            return
        if len(stack) > 1:
            _ = stack.pop(-1)
            mpd = stack[-1]

    def handle_character_data(texts: str):
        if texts.strip() != '':
            mpd.innertext += texts.strip()
    stack = []
    mpd = None # type: MPD
    mpd_handlers = {
        'MPD': MPD,
        'Location': Location,
        'BaseURL': BaseURL,
        'Period': Period,
        'AdaptationSet': AdaptationSet,
        'Representation': Representation,
        'SegmentTemplate': SegmentTemplate,
        'SegmentURL': SegmentURL,
        'SegmentBase': SegmentBaee,
        'Initialization': Initialization,
        'SegmentList': SegmentList,
        'SegmentTimeline': SegmentTimeline,
        'Role': Role,
        'S': S,
        'ContentProtection': ContentProtection,
        'cenc:pssh': CencPssh,
    }
    parser = ParserCreate()
    parser.StartElementHandler = handle_start_element
    parser.EndElementHandler = handle_end_element
    parser.CharacterDataHandler = handle_character_data
    parser.Parse(content)
    return mpd


class BaseUri:
    def __init__(self, name: str = '', home_url: str = '', base_url: str = ''):
        self.name = name
        self.home_url = home_url
        self.base_url = base_url

    def new_name(self, name: str):
        return BaseUri(name, self.home_url, self.base_url)

    def new_home_url(self, home_url: str):
        return BaseUri(self.name, home_url, self.base_url)

    def new_base_url(self, base_url: str):
        return BaseUri(self.name, self.home_url, base_url)


import re
from urllib.parse import urlparse
from hlsdl.models.segment import Segment


class DASHSegment(Segment):
    def __init__(self):
        super(DASHSegment, self).__init__()
        self.suffix = '.mp4'

    def is_encrypt(self):
        return True

    def is_supported_encryption(self):
        return False

    def set_duration(self, duration: float):
        self.duration = duration

    def set_fmt_time(self, fmt_time: float):
        self.fmt_time = fmt_time

    def set_subtitle_url(self, subtitle_url: str):
        self.name = subtitle_url.split('?')[0].split('/')[-1]
        self.index = -1
        self.url = subtitle_url
        self.segment_type = 'init'

    def set_init_url(self, init_url: str):
        parts = init_url.split('?')[0].split('/')[-1].split('.')
        if len(parts) > 1:
            self.suffix = f'.{parts[-1]}'
        self.name = f'init{self.suffix}'
        self.index = -1
        self.url = init_url
        self.segment_type = 'init'

    def get_url_name(self, url: str):
        url_name = urlparse(url).path.split('/')[-1]
        match = re.match('Fragments\((.+?)\)', url_name)
        if match:
            url_name = match.group(1).split(',')[0].replace('=', '_') + self.suffix
        return url_name

    def set_media_url(self, media_url: str, name_from_url: bool = False):
        parts = media_url.split('?')[0].split('/')[-1].split('.')
        if len(parts) > 1:
            # 修正后缀
            self.suffix = f'.{parts[-1]}'
            self.name = f'{self.index:0>4}.{parts[-1]}'
        self.url = media_url
        if name_from_url:
            self.name = self.get_url_name(self.url)


import re
from hlsdl.models.segment import Segment
from hlsdl.extractors.hls.ext.xkey import XKey
from hlsdl.extractors.hls.ext.xprivinf import XPrivinf


class HLSSegment(Segment):
    def __init__(self):
        super(HLSSegment, self).__init__()
        # 加密信息
        self.xkey = None # type: XKey
        self.__xprivinf = None # type: XPrivinf
        self.has_set_key = False

    def is_encrypt(self):
        if self.__xprivinf is not None:
            return self.__xprivinf.drm_notencrypt
        elif self.xkey is not None:
            return True
        else:
            return False

    def is_supported_encryption(self):
        if self.xkey is not None and self.xkey.method.upper() in ['AES-128']:
            return True
        return False

    def set_duration(self, line: str):
        try:
            self.duration = float(line.split(':', maxsplit=1)[-1].split(',')[0])
        except Exception:
            pass

    def set_byterange(self, line: str):
        try:
            _ = line.split(':', maxsplit=1)[-1].split('@')
            total, offset = int(_[0]), int(_[1])
            self.byterange = [total, offset]
        except Exception:
            pass

    def set_privinf(self, line: str):
        '''
        对于分段来说 标签的属性值 应该归属在标签下面 计算时需要注意
        不过也可以在解析标签信息之后 进行赋值处理 这样便于调用
        '''
        self.__xprivinf = XPrivinf().set_attrs_from_line(line)
        if self.__xprivinf.filesize is not None:
            self.filesize = self.__xprivinf.filesize

    def set_url(self, home_url: str, base_url: str, line: str):
        if line.startswith('http://') or line.startswith('https://') or line.startswith('ftp://'):
            self.url = line
        elif line.startswith('/'):
            self.url = f'{home_url}/{line}'
        else:
            self.url = f'{base_url}/{line}'

    def set_map_url(self, home_url: str, base_url: str, line: str):
        map_uri = re.match('#EXT-X-MAP:URI="(.*?)"', line.strip())
        if map_uri is None:
            print('find #EXT-X-MAP tag, however has no uri')
            return
        map_uri = map_uri.group(1)
        if map_uri.startswith('http://') or map_uri.startswith('https://') or map_uri.startswith('ftp://'):
            self.url = map_uri
        elif map_uri.startswith('/'):
            self.url = f'{home_url}{map_uri}'
        else:
            self.url = f'{base_url}/{map_uri}'
        self.segment_type = 'map'
        # 每一条流理应只有一个map
        self.name = 'map.mp4'
        self.index = -1

    def set_key(self, home_url: str, base_url: str, line: str):
        self.has_set_key = True
        xkey = XKey().set_attrs_from_line(home_url, base_url, line)
        if xkey is not None:
            self.xkey = xkey

    def get_xkey(self):
        return self.xkey

    def set_xkey(self, last_segment_has_xkey: bool, xkey: XKey):
        '''
        如果已经因为#EXT-X-KEY而设置过xkey了
        那就不使用之前分段的xkey了
        '''
        if last_segment_has_xkey is False:
            return
        if self.has_set_key is True:
            return
        if xkey is None:
            return
        self.xkey = xkey



from typing import List
from hlsdl.models.key import StreamKey
from .childs.contentprotection import ContentProtection
from .childs.cencpssh import CencPssh


class COMMON_MPEG:
    def __init__(self, schemeiduri: str, cenc_default_kid: str, value: str):
        self.schemeiduri = schemeiduri
        self.cenc_default_kid = cenc_default_kid
        self.value = value


class COMMON_CENC:
    def __init__(self, schemeiduri: str, cenc_pssh: str, value: str):
        self.schemeiduri = schemeiduri
        self.cenc_pssh = cenc_pssh
        self.value = value


class MARLIN:
    def __init__(self, schemeiduri: str, mas_marlincontentids: list):
        self.schemeiduri = schemeiduri
        self.mas_marlincontentids = mas_marlincontentids


class PLAYREADY:
    def __init__(self, schemeiduri: str, mspr_pro: str):
        self.schemeiduri = schemeiduri
        self.mspr_pro = mspr_pro


class WIDEVINE:
    def __init__(self, schemeiduri: str, cenc_pssh: str):
        self.schemeiduri = schemeiduri
        self.cenc_pssh = cenc_pssh


class PRIMETIME:
    def __init__(self, schemeiduri: str, cenc_pssh: str):
        self.schemeiduri = schemeiduri
        self.cenc_pssh = cenc_pssh


class DASHKey(StreamKey):
    def __init__(self, cp: ContentProtection):
        super(DASHKey, self).__init__()
        key = ''
        method = ''
        # https://dashif.org/identifiers/content_protection/
        if cp.schemeIdUri == 'urn:mpeg:dash:mp4protection:2011':
            key = COMMON_MPEG(cp.schemeIdUri, cp.cenc_default_KID, cp.value)
            method = 'COMMON_MPEG'
        elif cp.schemeIdUri == 'urn:uuid:1077efec-c0b2-4d02-ace3-3c1e52e2fb4b':
            key = COMMON_CENC(cp.schemeIdUri, self.get_pssh(cp), cp.value)
            method = 'COMMON_CENC'
        elif cp.schemeIdUri == 'urn:uuid:5E629AF5-38DA-4063-8977-97FFBD9902D4':
            key = MARLIN(cp.schemeIdUri, [])
            method = 'MARLIN'
        elif cp.schemeIdUri == 'urn:uuid:9a04f079-9840-4286-ab92-e65be0885f95':
            mspr_pros = cp.find('mspr:pros') # type: List[CencPssh]
            if len(mspr_pros) > 0:
                mspr_pro = mspr_pros[0].innertext
            else:
                mspr_pro = ''
            key = PLAYREADY(cp.schemeIdUri, mspr_pro)
            method = 'PLAYREADY'
        elif cp.schemeIdUri == 'urn:uuid:edef8ba9-79d6-4ace-a3c8-27dcd51d21ed':
            key = WIDEVINE(cp.schemeIdUri, self.get_pssh(cp))
            method = 'WIDEVINE'
        elif cp.schemeIdUri == 'urn:uuid:F239E769-EFA3-4850-9C16-A903C6932EFB':
            key = '???????'
            method = 'PRIMETIME'
        self.method = method
        self.key = key

    def get_pssh(self, cp: ContentProtection) -> str:
        cenc_psshs = cp.find('cenc:pssh') # type: List[CencPssh]
        if len(cenc_psshs) > 0:
            return cenc_psshs[0].innertext
        else:
            return ''


import re
from .x import X


class XMedia(X):
    '''
    #EXT-X-MEDIA 外挂媒体
    - TYPE=AUDIO,URI="",GROUP-ID="default-audio-group",NAME="stream_0",AUTOSELECT=YES,CHANNELS="2"
    '''
    def __init__(self):
        super(XMedia, self).__init__('#EXT-X-MEDIA')
        self.type = None # type: str
        self.uri = None # type: str
        self.group_id = None # type: str
        self.language = None # type: str
        self.assoc_language = None # type: str
        self.name = None # type: str
        self.default = None # type: str
        self.autoselect = None # type: str
        self.forced = None # type: str
        self.instream_id = None # type: str
        self.subtitles = None # type: str
        self.channels = None # type: int
        self.known_attrs = {
            'TYPE': 'type',
            'URI': 'uri',
            'GROUP-ID': 'group_id',
            'LANGUAGE': 'language',
            'ASSOC-LANGUAGE': 'assoc_language',
            'NAME': 'name',
            'DEFAULT': 'default',
            'AUTOSELECT': 'autoselect',
            'FORCED': 'forced',
            'INSTREAM-ID': 'instream_id',
            'CHARACTERISTICS': 'subtitles',
            'CHANNELS': int,
        }

    def convert_type(self, name: str, value: str, _type: type):
        if name == 'CHANNELS':
            try:
                value = re.findall('(\d+)', value)[0]
            except Exception:
                pass
        self.__setattr__(self.format_key(name), _type(value))

    def set_attrs_from_line(self, line: str):
        '''
        这里实际上可以不写
        '''
        return super(XMedia, self).set_attrs_from_line(line)


import sys
import json
from pathlib import Path
from hlsdl.cmdargs import CmdArgs
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class Headers:
    def __init__(self):
        self.headers = {}

    def get(self, args: CmdArgs) -> dict:
        if getattr(sys, 'frozen', False):
            config_path = Path(sys.executable).parent / args.headers
        else:
            config_path = Path(__file__).parent.parent.parent / args.headers
        if config_path.exists() is False:
            logger.warning(f'{config_path.stem} is not exists, put your config file to {config_path.parent.parent.resolve().as_posix()}')
            return
        try:
            self.headers = json.loads(config_path.read_text(encoding='utf-8'))
        except Exception as e:
            logger.error(f'try to load {config_path.resolve().as_posix()} failed', exc_info=e)
        logger.debug(f'use headers:\n{json.dumps(self.headers, ensure_ascii=False, indent=4)}')
        return self.headers


from datetime import datetime
from .x import X


class XProgramDateTime(X):
    '''
    #EXT-X-PROGRAM-DATE-TIME 第一个分段的绝对时间
    - 2019-01-01T00:00:00.000Z
    '''
    def __init__(self):
        super(XProgramDateTime, self).__init__('#EXT-X-PROGRAM-DATE-TIME')
        self.program_date_time = None # type: datetime

    def set_attrs_from_line(self, line: str):
        '''
        重写父类同名函数
        '''
        line = self.get_tag_info(line)
        if line.endswith('Z') is True:
            line = f'{line[:-1]}+00:00'
        try:
            self.program_date_time = datetime.fromisoformat(line)
        except Exception:
            raise
        return self


import io
import struct

u8 = struct.Struct('>B')
u88 = struct.Struct('>Bx')
u16 = struct.Struct('>H')
u1616 = struct.Struct('>Hxx')
u32 = struct.Struct('>I')
u64 = struct.Struct('>Q')

s88 = struct.Struct('>bx')
s16 = struct.Struct('>h')
s1616 = struct.Struct('>hxx')
s32 = struct.Struct('>i')
unity_matrix = (s32.pack(0x10000) + s32.pack(0) * 3) * 2 + s32.pack(0x40000000)


def box(box_type, payload):
    return u32.pack(8 + len(payload)) + box_type + payload


def full_box(box_type, version, flags, payload):
    return box(box_type, u8.pack(version) + u32.pack(flags)[1:] + payload)


def extract_box_data(data: bytes, box_sequence: list):
    data_reader = io.BytesIO(data)
    while True:
        box_size = u32.unpack(data_reader.read(4))[0]
        box_type = data_reader.read(4)
        if box_type == box_sequence[0]:
            box_data = data_reader.read(box_size - 8)
            if len(box_sequence) == 1:
                return box_data
            return extract_box_data(box_data, box_sequence[1:])
        data_reader.seek(box_size - 8, 1)


class MetaItem:
    def generate(self):
        pass

    def to_int(self):
        pass

    def match_duration(self, _duration: str) -> float:
        '''
        test samples
        - PT50M0S
        - PT1H54.600S
        - PT23M59.972S
        - P8DT11H6M41.1367016S
        - P0Y0M0DT0H3M30.000S
        '''
        if isinstance(_duration, str) is False:
            return

        def reset_token():
            nonlocal token_unit, token_time
            token_unit, token_time = '', ''
        offset = 0
        duration = 0.0
        token_unit = ''
        token_time = ''
        t_flag = False
        while offset < len(_duration):
            if _duration[offset].isalpha():
                token_unit += _duration[offset]
            elif _duration[offset].isdigit() or _duration[offset] == '.':
                token_time += _duration[offset]
            else:
                assert False, f'not possible be here _duration => {_duration}'
            offset += 1
            if token_unit == 'P':
                reset_token()
            elif token_unit == 'Y' or (t_flag is False and token_unit == 'M'):
                # 暂时先不计算年和月 有问题再说
                reset_token()
            elif token_unit == 'D':
                duration += 24 * int(token_time) * 60 * 60
                reset_token()
            elif token_unit == 'T':
                t_flag = True
                reset_token()
            elif token_unit == 'H':
                duration += int(token_time) * 60 * 60
                reset_token()
            elif token_unit == 'M':
                duration += int(token_time) * 60
                reset_token()
            elif token_unit == 'S':
                duration += float("0" + token_time)
                reset_token()
        return duration


from typing import List
from pathlib import Path


class Segment:
    '''
    每一个分段应当具有以下基本属性：
    - 名称
    - 索引
    - 后缀
    - 链接
    - 文件大小
    - 时长
    - 二进制内容
    - 下载文件夹
    - 分段类型
    '''
    def __init__(self):
        self.name = ''
        self.index = 0
        self.suffix = '.ts'
        self.url = ''
        self.filesize = 0
        self.duration = 0.0
        # dash直播流需要通过比较时间来确定是不是需要下载
        self.fmt_time = 0
        self.byterange = [] # type: list
        # <---临时存放二进制内容--->
        self.content = [] # type: List[bytes]
        # <---分段临时下载文件夹--->
        self.folder = None # type: Path
        # <---分段类型--->
        self.segment_type = 'normal'
        self.skip_concat = False
        # 直播流 单个分段 最大404次数
        self.max_retry_404 = 5

    def is_ism(self) -> bool:
        ''' 请重写 '''
        return False

    def is_encrypt(self) -> bool:
        ''' 请重写 '''
        pass

    def is_supported_encryption(self) -> bool:
        ''' 请重写 '''
        pass

    def add_offset_for_name(self, offset: int, has_init: bool = False, name_from_url: bool = False):
        self.index += offset
        if has_init:
            self.index -= 1
        if name_from_url is False:
            self.name = f'{self.index:0>4}{self.suffix}'

    def set_offset_for_name(self, offset: int, has_init: bool = False, name_from_url: bool = False):
        self.index = offset
        if has_init:
            self.index -= 1
        if name_from_url is False:
            self.name = f'{self.index:0>4}{self.suffix}'

    def set_index(self, index: str):
        self.index = index
        if index == -1:
            self.name = f'init{self.suffix}'
        else:
            self.name = f'{self.index:0>4}{self.suffix}'
        return self

    def set_folder(self, folder: Path):
        self.folder = folder
        return self

    def get_path(self) -> Path:
        return self.folder / self.name

    def dump(self) -> bool:
        self.get_path().write_bytes(b''.join(self.content))
        self.content = []
        return True


import locale


class Texts:
    def __init__(self):
        if locale.getdefaultlocale()[0] == 'zh_CN':
            self.setup_zh_CN()
        else:
            self.setup_en_US()

    def setup_zh_CN(self):
        self.input_stream_number = '请输入要下载流的序号：\n'
        self.select_without_any_stream = '未选择流，退出'
        self.download_start = '下载开始'
        self.segment_cannot_download = '无法下载的m3u8 取消全部下载任务\n'
        self.segment_cannot_download_unknown_status = '出现未知status 取消全部下载任务\n'
        self.segment_cannot_download_unknown_exc = '出现未知异常 强制退出全部task'
        self.cannot_get_stream_metadata = '无法获取视频流信息'
        self.decrypted_file_exists_skip = '解密文件已存在 跳过'
        self.start_decrypt = '开始解密'
        self.total_segments_info_1 = '共计'
        self.total_segments_info_2 = '个分段'
        self.try_to_concat = '尝试合并'
        self.cancel_concat_reason_1 = '但是已经存在合并文件'
        self.cancel_concat_reason_2 = '但是未下载完成'
        self.force_use_raw_concat_for_sample_aes = '发现SAMPLE-AES 将使用二进制合并'

    def setup_en_US(self):
        self.input_stream_number = 'input stream number(s):\n'
        self.select_without_any_stream = 'haven\'t select any stream, exiting'
        self.download_start = 'download start'
        self.segment_cannot_download = 'can not download segment, cancel all downloading task\n'
        self.segment_cannot_download_unknown_status = 'appear unknown status, cancel all downloading task\n'
        self.segment_cannot_download_unknown_exc = 'appear unknown status, force to cancel all downloading task'
        self.cannot_get_stream_metadata = 'cannot get stream metadata'
        self.decrypted_file_exists_skip = 'decrypted file exists, skip it'
        self.start_decrypt = 'start decrypt'
        self.total_segments_info_1 = 'total'
        self.total_segments_info_2 = ' segments'
        self.try_to_concat = 'try to concat'
        self.cancel_concat_reason_1 = 'but file already exists'
        self.cancel_concat_reason_2 = 'buf download not completely'
        self.force_use_raw_concat_for_sample_aes = 'force use --raw-concat for SAMPLE-AES(-CTR)'


t_msg = Texts()


# if __name__ == '__main__':
#     t_msg = Texts()
#     print(t_msg.try_to_concat)


from .ismitem import ISMItem


class ISM(ISMItem):
    def __init__(self, name: str):
        super(ISM, self).__init__(name)
        self.MajorVersion = None # type: str
        self.MinorVersion = None # type: str
        self.TimeScale = None # type: int
        self.Duration = None # type: int

    def generate(self):
        self.to_int('TimeScale')
        self.to_int('Duration')


import re
import time
from typing import List
from pathlib import Path
from hlsdl.models.stream import Stream
from hlsdl.models.base import BaseUri
from hlsdl.extractors.mss.segment import MSSSegment
from hlsdl.extractors.mss.box_util import (
    u8,
    u16,
    u1616,
    u32,
    u64,
    s88,
    s16,
    s1616,
    box,
    full_box,
    unity_matrix,
    extract_box_data,
)

TRACK_ENABLED = 0x1
TRACK_IN_MOVIE = 0x2
TRACK_IN_PREVIEW = 0x4

SELF_CONTAINED = 0x1

NALUTYPE_SPS = 7
NALUTYPE_PPS = 8


class MSSStream(Stream):
    def __init__(self, index: int, uri_item: BaseUri, save_dir: Path):
        super(MSSStream, self).__init__(index, uri_item, save_dir)
        self.model = 'mss'
        self.timescale = 10000000
        self.channels = 2
        self.bits_per_sample = 16
        self.sampling_rate = 16
        self.codec_private_data = None # type: str
        self.nal_unit_length_field = 4
        self.segments = [] # type: List[MSSSegment]
        self.suffix = '.mp4'
        self.has_init_segment = False
        self.skey = '' # type: str
        self.kid = bytes([0] * 16)
        self.track_index = 0
        self.width = self.height = 0
        self.set_init_segment()
        self.append_segment()

    def set_kid(self, kid: bytes):
        self.kid = kid

    def get_name(self):
        if self.stream_type != '':
            base_name = f'{self.name}_{self.stream_type}'
        else:
            base_name = self.name
        if self.codecs is not None:
            base_name += f'_{self.codecs}'
        if self.stream_type == 'text' and self.lang != '':
            base_name += f'_{self.lang}'
        elif self.stream_type == 'video' and self.resolution != '':
            base_name += f'_{self.resolution}'
        elif self.stream_type == 'audio' and self.lang != '':
            base_name += f'_{self.lang}'
        if self.stream_type in ['audio', 'video'] and self.bandwidth is not None:
            base_name += f'_{self.bandwidth / 1000:.2f}kbps'
        return base_name

    def get_ism_params(self):
        height_width = self.resolution.split('x')
        if len(height_width) == 2:
            height, width = height_width
        else:
            height, width = 0, 0
        return {
            'fourcc': self.codecs,
            'duration': int(self.duration),
            'timescale': self.timescale,
            'language': 'und' if self.lang == '' else self.lang,
            'height': int(height),
            'width': int(width),
            'stream_type': self.stream_type,
            'channels': self.channels,
            'bits_per_sample': self.bits_per_sample,
            'sampling_rate': self.sampling_rate,
            'codec_private_data': self.codec_private_data,
            'nal_unit_length_field': self.nal_unit_length_field,
        }

    def set_init_segment(self):
        self.has_init_segment = True
        segment = MSSSegment().set_index(-1).set_folder(self.save_dir)
        self.segments.append(segment)

    def append_segment(self):
        index = len(self.segments)
        if self.has_init_segment:
            index -= 1
        segment = MSSSegment().set_index(index).set_folder(self.save_dir)
        self.segments.append(segment)

    def set_subtitle_url(self, url: str):
        self.has_init_segment = True
        self.segments[-1].set_subtitle_url(self.fix_url(url))
        # self.append_segment()

    def set_init_url(self, url: str):
        self.has_init_segment = True
        self.segments[-1].set_init_url(self.fix_url(url))
        self.append_segment()

    def set_media_url(self, url: str):
        self.segments[-1].set_media_url(self.fix_url(url))
        self.append_segment()

    def set_segment_duration(self, duration: float):
        self.segments[-1].set_duration(duration)

    def set_segments_duration(self, duration: float):
        '''' init分段没有时长 这里只用设置普通分段的 '''
        for segment in self.segments:
            segment.set_duration(duration)

    def set_protection_flag(self, flag: bool):
        for segment in self.segments:
            segment.set_protection_flag(flag)

    def set_lang(self, lang: str):
        if lang is None:
            return
        self.lang = lang.lower()

    def set_bandwidth(self, bandwidth: int):
        self.bandwidth = bandwidth

    def set_codecs(self, codecs: str):
        if re.match('avc(1|3)*', codecs.lower()):
            codecs = 'H264'
        if re.match('(hev|hvc)1*', codecs.lower()):
            codecs = 'H265'
        if re.match('vp(09|9)*', codecs.lower()):
            codecs = 'VP9'
        if codecs != 'AACL' and re.match('aac*', codecs.lower()):
            codecs = 'AAC'
        if codecs.lower() in ['wvtt', 'ttml']:
            codecs = codecs.upper()
        self.codecs = codecs

    def set_resolution(self, width: str, height: str):
        if width is None or height is None:
            return
        self.width = int(width)
        self.height = int(height)
        self.resolution = f'{width}x{height}'

    def set_stream_type(self, stream_type: str):
        if stream_type is None:
            return
        self.stream_type = stream_type
        if self.stream_type == 'audio':
            self.suffix = '.m4a'

    def set_track_index(self, track_index: int):
        ''' to calc track id '''
        self.track_index = track_index

    def get_track_name(self):
        if self.track_name:
            return f'{self.track_name}_{self.track_index}'
        else:
            return f'{self.stream_type}_{self.track_index}'

    def set_track_name(self, track_name: str):
        self.track_name = track_name

    def set_timescale(self, timescale: int):
        if timescale is None:
            return
        self.timescale = timescale

    def set_bits_per_sample(self, bits_per_sample: str):
        if bits_per_sample is None:
            return
        self.bits_per_sample = bits_per_sample

    def set_sampling_rate(self, sampling_rate: str):
        if sampling_rate is None:
            return
        self.sampling_rate = sampling_rate

    def set_channels(self, channels: str):
        if channels is None:
            return
        self.channels = channels

    def set_codec_private_data(self, codec_private_data: str):
        if codec_private_data is None:
            return
        self.codec_private_data = codec_private_data

    def set_nal_unit_length_field(self, nal_unit_length_field: int):
        if nal_unit_length_field is None:
            return
        self.nal_unit_length_field = nal_unit_length_field

    def fix_header(self, is_fake: bool):
        if is_fake:
            track_id = 1
        else:
            tfhd_data = extract_box_data(self.segments[1].get_path().read_bytes(), [b'moof', b'traf', b'tfhd'])
            track_id = u32.unpack(tfhd_data[4:8])[0]
        init_payload = self.write_iso6_header(track_id, is_enc=self.segments[0].has_protection)
        self.segments[0].get_path().write_bytes(init_payload)

    def get_sinf_payload(self, kid: bytes, codec: bytes):
        sinf_payload = box(b'frma', codec)

        schm_payload = u32.pack(0x63656E63) # scheme_type 'cenc' => common encryption
        schm_payload += u32.pack(0x00010000) # scheme_version Major version 1, Minor version 0
        sinf_payload += full_box(b'schm', 0, 0, schm_payload)

        tenc_payload = u8.pack(0x0) * 2
        tenc_payload += u8.pack(0x1) # default_IsEncrypted
        tenc_payload += u8.pack(0x8) # default_IV_size
        tenc_payload += kid # default_KID
        tenc_payload = full_box(b'tenc', 0, 0, tenc_payload)
        sinf_payload += box(b'schi', tenc_payload)
        sinf_payload = box(b'sinf', sinf_payload)
        return sinf_payload

    def write_iso6_header(self, track_id: int, write_time: bool = False, is_enc: bool = False):
        '''
        根据dashif的信息，需要以下信息构成ism的init数据
        - track_name
        - track_id
        - FourCC
        - duration
        - timescale
        - language
        - height 仅视频
        - width 仅视频
        - bandwidth
        - kid
        - 是否加密
        - 类型
        '''
        if write_time:
            creation_time = modification_time = int(time.time())
        else:
            creation_time = modification_time = 0
        track_name = self.get_track_name()
        fourcc = self.codecs
        duration = self.duration
        timescale = self.timescale
        language = 'und' if self.lang == '' else self.lang
        height = self.height
        width = self.width
        bandwidth = self.bandwidth
        kid = self.kid
        stream_type = self.stream_type
        channels = self.channels
        bits_per_sample = self.bits_per_sample
        sampling_rate = self.sampling_rate
        codec_private_data = self.codec_private_data

        # ftyp box
        ftyp_payload = b'iso6' # major brand
        ftyp_payload += u32.pack(1) # minor version
        ftyp_payload += b'isom' + b'iso6' + b'msdh' # compatible brands

        mvhd_payload = u64.pack(creation_time)
        mvhd_payload += u64.pack(modification_time)
        mvhd_payload += u32.pack(timescale)
        mvhd_payload += u64.pack(int(duration * timescale))
        mvhd_payload += s1616.pack(1) # rate
        mvhd_payload += s88.pack(1) # volume
        mvhd_payload += u16.pack(0) # reserved1
        mvhd_payload += u32.pack(0) * 2 # reserved2
        mvhd_payload += unity_matrix
        mvhd_payload += u32.pack(0) * 6 # pre defined
        # 需要确认是不是要 +1
        mvhd_payload += u32.pack(track_id + 1) # next track id
        moov_payload = full_box(b'mvhd', 1, 0, mvhd_payload) # Movie Header Box

        tkhd_payload = u64.pack(creation_time)
        tkhd_payload += u64.pack(modification_time)
        tkhd_payload += u32.pack(track_id) # track id
        tkhd_payload += u32.pack(0) # reserved1
        tkhd_payload += u64.pack(int(duration * timescale))
        tkhd_payload += u32.pack(0) * 2 # reserved2
        tkhd_payload += s16.pack(0) # layer
        tkhd_payload += s16.pack(0) # alternate group
        # 1.0 ???
        tkhd_payload += s88.pack(1) # volume
        tkhd_payload += u16.pack(0) # reserved3
        tkhd_payload += unity_matrix
        tkhd_payload += u1616.pack(width)
        tkhd_payload += u1616.pack(height)
        trak_payload = full_box(b'tkhd', 1, TRACK_ENABLED | TRACK_IN_MOVIE | TRACK_IN_PREVIEW, tkhd_payload) # Track Header Box

        mdhd_payload = u64.pack(creation_time)
        mdhd_payload += u64.pack(modification_time)
        mdhd_payload += u32.pack(timescale)
        mdhd_payload += u64.pack(int(duration * timescale))
        mdhd_payload += u16.pack(((ord(language[0]) - 0x60) << 10) | ((ord(language[1]) - 0x60) << 5) | (ord(language[2]) - 0x60))
        mdhd_payload += u16.pack(0) # pre defined
        mdia_payload = full_box(b'mdhd', 1, 0, mdhd_payload) # Media Header Box

        hdlr_payload = u32.pack(0) # pre defined
        if stream_type == 'video': # handler type
            hdlr_payload += b'vide'
        elif stream_type == 'audio':
            hdlr_payload += b'soun'
        elif stream_type == 'text':
            hdlr_payload += b'subt'
        else:
            hdlr_payload += b'meta'
            assert False
        hdlr_payload += u32.pack(0) * 3 # reserved
        hdlr_payload += track_name.encode('utf-8') + b'\0' # name
        mdia_payload += full_box(b'hdlr', 0, 0, hdlr_payload) # Handler Reference Box

        if stream_type == 'video':
            vmhd_payload = u16.pack(0) # graphics mode
            vmhd_payload += u16.pack(0) * 3 # opcolor
            media_header_box = full_box(b'vmhd', 0, 1, vmhd_payload) # Video Media Header
        elif stream_type == 'audio':
            smhd_payload = s88.pack(0) # balance
            smhd_payload += u16.pack(0) # reserved
            media_header_box = full_box(b'smhd', 0, 1, smhd_payload) # Sound Media Header
        elif stream_type == 'text':
            media_header_box = full_box(b'sthd', 0, 1, b'') # Subtitle Media Header
        else:
            assert False
        minf_payload = media_header_box

        dref_payload = u32.pack(1) # entry count
        dref_payload += full_box(b'url ', 0, SELF_CONTAINED, b'') # Data Entry URL Box
        dinf_payload = full_box(b'dref', 0, 0, dref_payload) # Data Reference Box
        minf_payload += box(b'dinf', dinf_payload) # Data Information Box

        stsd_payload = u32.pack(1) # entry count

        sample_entry_payload = u8.pack(0) * 6 # reserved1
        sample_entry_payload += u16.pack(1) # data reference index
        if stream_type == 'audio':
            sample_entry_payload += u32.pack(0) * 2 # reserved2
            sample_entry_payload += u16.pack(channels)
            sample_entry_payload += u16.pack(bits_per_sample)
            sample_entry_payload += u16.pack(0) # pre defined
            sample_entry_payload += u16.pack(0) # reserved3
            # 不管这里多少 mediainfo 查看出来的都是正确的数值 可能通过其他部分计算
            sample_entry_payload += u1616.pack(sampling_rate)

            audioSpecificConfig = bytes.fromhex(codec_private_data)

            # ESDS length = esds box header length (= 12) +
            #               ES_Descriptor header length (= 5) +
            #               DecoderConfigDescriptor header length (= 15) +
            #               decoderSpecificInfo header length (= 2) +
            #               AudioSpecificConfig length (= codecPrivateData length)
            # esdsLength = 34 + len(audioSpecificConfig)

            # ES_Descriptor (see ISO/IEC 14496-1 (Systems))
            esds_payload = u8.pack(0x03) # tag = 0x03 (ES_DescrTag)
            esds_payload += u8.pack(20 + len(audioSpecificConfig)) # size
            esds_payload += u8.pack((track_id & 0xFF00) >> 8) # ES_ID = track_id
            esds_payload += u8.pack(track_id & 0x00FF)
            esds_payload += u8.pack(0) # flags and streamPriority

            # DecoderConfigDescriptor (see ISO/IEC 14496-1 (Systems))
            esds_payload += u8.pack(0x04) # tag = 0x04 (DecoderConfigDescrTag)
            esds_payload += u8.pack(15 + len(audioSpecificConfig)) # size
            esds_payload += u8.pack(0x40) # objectTypeIndication = 0x40 (MPEG-4 AAC)
            # esds_payload[i] = 0x05 << 2 # streamType = 0x05 (Audiostream)
            # esds_payload[i] |= 0 << 1 # upStream = 0
            esds_payload += u8.pack((0x05 << 2) | (0 << 1) | 1) # reserved = 1
            esds_payload += u8.pack(0xFF) # buffersizeDB = undefined
            esds_payload += u8.pack(0xFF) # ''
            esds_payload += u8.pack(0xFF) # ''
            esds_payload += u8.pack((bandwidth & 0xFF000000) >> 24) # maxBitrate
            esds_payload += u8.pack((bandwidth & 0x00FF0000) >> 16) # ''
            esds_payload += u8.pack((bandwidth & 0x0000FF00) >> 8) # ''
            esds_payload += u8.pack((bandwidth & 0x000000FF)) # ''
            esds_payload += u8.pack((bandwidth & 0xFF000000) >> 24) # avgbitrate
            esds_payload += u8.pack((bandwidth & 0x00FF0000) >> 16) # ''
            esds_payload += u8.pack((bandwidth & 0x0000FF00) >> 8) # ''
            esds_payload += u8.pack((bandwidth & 0x000000FF)) # ''

            # DecoderSpecificInfo (see ISO/IEC 14496-1 (Systems))
            esds_payload += u8.pack(0x05) # tag = 0x05 (DecSpecificInfoTag)
            esds_payload += u8.pack(len(audioSpecificConfig)) # size
            esds_payload += audioSpecificConfig # AudioSpecificConfig bytes
            if is_enc:
                sample_entry_payload += full_box(b'esds', 0, 0, esds_payload) + self.get_sinf_payload(kid, b'mp4a')
                sample_entry_box = box(b'enca', sample_entry_payload)
            else:
                sample_entry_payload += full_box(b'esds', 0, 0, esds_payload)
                sample_entry_box = box(b'mp4a', sample_entry_payload)
        elif stream_type == 'video':
            sample_entry_payload += u16.pack(0) # pre defined
            sample_entry_payload += u16.pack(0) # reserved
            sample_entry_payload += u32.pack(0) * 3 # pre defined
            sample_entry_payload += u16.pack(width)
            sample_entry_payload += u16.pack(height)
            sample_entry_payload += u1616.pack(0x48) # horiz resolution 72 dpi
            sample_entry_payload += u1616.pack(0x48) # vert resolution 72 dpi
            sample_entry_payload += u32.pack(0) # reserved
            sample_entry_payload += u16.pack(1) # frame count
            # sample_entry_payload += u8.pack(0) * 32 # compressor name
            sample_entry_payload += bytes([
                0x0A, 0x41, 0x56, 0x43, 0x20, 0x43, 0x6F, 0x64,
                0x69, 0x6E, 0x67, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
            ]) # compressor name
            sample_entry_payload += u16.pack(0x18) # depth
            sample_entry_payload += u16.pack(65535) # pre defined3

            if fourcc in ('H264', 'AVC1'):
                # avcCLength = 15
                sps = []
                pps = []
                nalus = codec_private_data.split('00000001')[1:]
                for nalu in nalus:
                    naluBytes = bytes.fromhex(nalu)
                    naluType = naluBytes[0] & 0x1F
                    if naluType == NALUTYPE_SPS:
                        sps.append(naluBytes)
                        # avcCLength += len(naluBytes) + 2 # 2 = sequenceParameterSetLength field length
                    elif naluType == NALUTYPE_PPS:
                        # avcCLength += len(naluBytes) + 2 # 2 = pictureParameterSetLength field length
                        pps.append(naluBytes)
                    else:
                        pass
                if len(sps) > 0:
                    AVCProfileIndication = sps[0][1]
                    profile_compatibility = sps[0][2]
                    AVCLevelIndication = sps[0][3]
                else:
                    AVCProfileIndication = 0
                    profile_compatibility = 0
                    AVCLevelIndication = 0
                # 这里不知道为什么没用上
                # avcc_head_payload = bytes([
                #     (avcCLength & 0xFF000000) >> 24,
                #     (avcCLength & 0x00FF0000) >> 16,
                #     (avcCLength & 0x0000FF00) >> 8,
                #     (avcCLength & 0x000000FF),
                # ])

                avcc_payload = u8.pack(1) # configurationVersion = 1
                avcc_payload += u8.pack(AVCProfileIndication)
                avcc_payload += u8.pack(profile_compatibility)
                avcc_payload += u8.pack(AVCLevelIndication)
                avcc_payload += u8.pack(0xFF) # '11111' + lengthSizeMinusOne = 3
                avcc_payload += u8.pack(0xE0 | len(sps)) # '111' + numOfSequenceParameterSets
                for item in sps:
                    avcc_payload += u8.pack((len(item) & 0xFF00) >> 8)
                    avcc_payload += u8.pack((len(item) & 0x00FF))
                    avcc_payload += item
                avcc_payload += u8.pack(len(pps)) # numOfPictureParameterSets
                for item in pps:
                    avcc_payload += u8.pack((len(item) & 0xFF00) >> 8)
                    avcc_payload += u8.pack((len(item) & 0x00FF))
                    avcc_payload += item

                sample_entry_payload += box(b'avcC', avcc_payload) + self.get_sinf_payload(kid, b'avc1') # AVC Decoder Configuration Record
                sample_entry_box = box(b'encv', sample_entry_payload) # AVC Simple Entry
            else:
                assert False
        elif stream_type == 'text':
            if fourcc == 'TTML':
                sample_entry_payload += b'http://www.w3.org/ns/ttml\0' # namespace
                sample_entry_payload += b'\0' # schema location
                sample_entry_payload += b'\0' # auxilary mime types(??)
                sample_entry_box = box(b'stpp', sample_entry_payload)
            else:
                assert False
        else:
            assert False

        stts_payload = u32.pack(0) # entry count
        stbl_payload = full_box(b'stts', 0, 0, stts_payload) # Decoding Time to Sample Box

        stsc_payload = u32.pack(0) # entry count
        stbl_payload += full_box(b'stsc', 0, 0, stsc_payload) # Sample To Chunk Box

        stco_payload = u32.pack(0) # entry count
        stbl_payload += full_box(b'stco', 0, 0, stco_payload) # Chunk Offset Box

        stsz_payload = u32.pack(0) + u32.pack(0) # sample size, sample count
        stbl_payload += full_box(b'stsz', 0, 0, stsz_payload) # Sample Size Box

        stsd_payload += sample_entry_box
        stbl_payload += full_box(b'stsd', 0, 0, stsd_payload) # Sample Description Box

        minf_payload += box(b'stbl', stbl_payload) # Sample Table Box
        mdia_payload += box(b'minf', minf_payload) # Media Information Box
        trak_payload += box(b'mdia', mdia_payload) # Media Box
        moov_payload += box(b'trak', trak_payload) # Track Box

        trex_payload = u32.pack(track_id) # track id
        trex_payload += u32.pack(1) # default sample description index
        trex_payload += u32.pack(0) # default sample duration
        trex_payload += u32.pack(0) # default sample size
        trex_payload += u32.pack(0) # default sample flags
        mvex_payload = full_box(b'trex', 0, 0, trex_payload) # Track Extends Box
        moov_payload += box(b'mvex', mvex_payload) # Movie Extends Box
        moov_payload = box(b'moov', moov_payload)
        moov_payload = box(b'ftyp', ftyp_payload) + moov_payload
        return moov_payload


from ..mpditem import MPDItem


class SegmentTemplate(MPDItem):
    '''
    SegmentTemplate没有duration的话 timescale好像没什么用
    '''
    def __init__(self, name: str):
        super(SegmentTemplate, self).__init__(name)
        self.timescale = 0 # type: int
        self.duration = 0 # type: int
        self.presentationTimeOffset = None # type: int
        self.initialization = None # type: str
        self.media = None # type: str
        self.startNumber = None # type: int

    def generate(self):
        if self.presentationTimeOffset is None:
            self.presentationTimeOffset = 0
        if self.startNumber is None:
            self.startNumber = 1
        self.to_int()

    def to_int(self):
        self.timescale = int(self.timescale)
        self.duration = int(self.duration)
        self.presentationTimeOffset = int(self.presentationTimeOffset)
        self.startNumber = int(self.startNumber)

    def get_url(self) -> str:
        return self.initialization
        # if self.initialization is None:
        #     return self.initialization
        # return self.initialization.replace('..', '')

    def get_media_url(self) -> str:
        return self.media
        # return self.media.replace('../', '')


from ..ismitem import ISMItem


class QualityLevel(ISMItem):
    '''
    https://docs.microsoft.com/en-us/iis/extensions/smooth-streaming-client/qualitylevel-attributes-iis-smooth-streaming
    '''
    def __init__(self, name: str):
        super(QualityLevel, self).__init__(name)
        # <----- 通用 ----->
        self.Index = None # type: int
        self.Bitrate = None # type: int
        self.CodecPrivateData = None # type: str
        self.FourCC = None # type: str
        self.NALUnitLengthField = None # type: int
        # <----- 音频 ----->
        self.SamplingRate = None # type: int
        self.Channels = None # type: int
        self.BitsPerSample = None # type: int
        self.PacketSize = None # type: int
        self.AudioTag = None # type: str
        # <----- 视频 ----->
        self.MaxWidth = None # type: str
        self.MaxHeight = None # type: str

    def generate(self):
        self.to_int('Index')
        self.to_int('Bitrate')
        self.to_int('SamplingRate')
        self.to_int('Channels')
        self.to_int('BitsPerSample')
        self.to_int('PacketSize')


from Crypto.Cipher import AES
from hlsdl.models.segment import Segment


class CommonAES:
    '''
    这是一个常规的AES-128-CBC解密类
    或许应该将这个类注册为分段的一个属性？
    '''

    def __init__(self, aes_key: bytes, aes_iv: bytes = None):
        self.aes_key = aes_key # type: bytes
        self.aes_iv = aes_iv # type: bytes
        if self.aes_iv is None:
            self.aes_iv = bytes([0] * 16)

    def decrypt(self, segment: Segment) -> bool:
        '''
        解密 落盘
        '''
        try:
            cipher = AES.new(self.aes_key, AES.MODE_CBC, iv=self.aes_iv)
            content = cipher.decrypt(b''.join(segment.content))
            segment.content = []
        except Exception as e:
            print(f'decrypt {segment.name} error -> {e}')
            return False
        else:
            segment.get_path().write_bytes(content)
            return True


from datetime import datetime
from .mpditem import MPDItem


class MPD(MPDItem):
    def __init__(self, name: str):
        super(MPD, self).__init__(name)
        self.maxSegmentDuration = None # type: str
        self.mediaPresentationDuration = None # type: str
        self.minBufferTime = None # type: str
        # live profile
        # - urn:mpeg:dash:profile:isoff-live:2011
        # - urn:mpeg:dash:profile:isoff-ext-live:2014
        self.profiles = None # type: str
        # dynamic -> live
        # static -> live playback
        self.type = None # type: str
        # only use when type is 'dynamic' which specifies the smallest period between potential changes to the MPD
        self.minimumUpdatePeriod = None # type: str
        # time of client to fetch the mpd content
        self.publishTime = None # type: datetime
        self.availabilityStartTime = None # type: float
        self.timeShiftBufferDepth = None # type: str
        self.suggestedPresentationDelay = None # type: str

    def generate(self):
        if isinstance(self.maxSegmentDuration, str):
            self.maxSegmentDuration = self.match_duration(self.maxSegmentDuration)
        if isinstance(self.mediaPresentationDuration, str):
            self.mediaPresentationDuration = self.match_duration(self.mediaPresentationDuration)
        if isinstance(self.minBufferTime, str):
            self.minBufferTime = self.match_duration(self.minBufferTime)
        if isinstance(self.minimumUpdatePeriod, str):
            self.minimumUpdatePeriod = self.match_duration(self.minimumUpdatePeriod)
        if isinstance(self.availabilityStartTime, str):
            # if self.availabilityStartTime in ['1970-01-01T00:00:00Z', '1970-01-01T00:00:00.000Z']:
            if self.availabilityStartTime.startswith('1970-01-01'):
                self.availabilityStartTime = 0.0
            # 2019-03-05T08:26:06.748000+00:00
            if isinstance(self.availabilityStartTime, str) and self.availabilityStartTime[-9:] == '000+00:00':
                self.availabilityStartTime = self.availabilityStartTime[:-9] + 'Z'
            try:
                self.availabilityStartTime = datetime.strptime(self.availabilityStartTime, '%Y-%m-%dT%H:%M:%S.%fZ').timestamp()
            except Exception:
                try:
                    self.availabilityStartTime = datetime.strptime(self.availabilityStartTime, '%Y-%m-%dT%H:%M:%SZ').timestamp()
                except Exception:
                    pass
        if isinstance(self.publishTime, str):
            is_match = False
            try:
                self.publishTime = datetime.strptime(self.publishTime, '%Y-%m-%dT%H:%M:%S.%fZ')
                is_match = True
            except Exception:
                pass
            if is_match is False:
                try:
                    self.publishTime = datetime.strptime(self.publishTime, '%Y-%m-%dT%H:%M:%SZ')
                    is_match = True
                except Exception:
                    pass
            if is_match is False:
                try:
                    # 2021-11-28T12:33:53
                    self.publishTime = datetime.strptime(self.publishTime, '%Y-%m-%dT%H:%M:%S')
                    is_match = True
                except Exception:
                    pass
            if is_match is False:
                try:
                    # 2022-02-14T11:43:04+00:00
                    self.publishTime = datetime.strptime(self.publishTime.split('+')[0], '%Y-%m-%dT%H:%M:%S')
                    is_match = True
                except Exception:
                    pass
            if is_match is False:
                assert is_match is True, f'match publishTime failed => {self.publishTime}'


from pathlib import Path


class CmdArgs:

    def __init__(self):
        self.speed_up = None # type: bool
        self.speed_up_left = None # type: int
        self.live = None # type: bool
        self.compare_with_url = None # type: bool
        self.dont_split_discontinuity = None # type: bool
        self.name_from_url = None # type: bool
        self.live_duration = None # type: float
        self.live_utc_offset = None # type: int
        self.live_refresh_interval = None # type: int
        self.name = None # type: str
        self.base_url = None # type: str
        self.ad_keyword = None # type: str
        self.resolution = None # type: str
        self.best_quality = None # type: bool
        self.video_only = None # type: bool
        self.audio_only = None # type: bool
        self.all_videos = None # type: bool
        self.all_audios = None # type: bool
        self.all_subtitles = None # type: bool
        self.service = None # type: str
        self.save_dir = None # type: Path
        self.ffmpeg = None # type: str
        self.mp4decrypt = None # type: str
        self.mp4box = None # type: str
        self.select = None # type: bool
        self.multi_s = None # type: bool
        self.disable_force_close = None # type: bool
        self.limit_per_host = None # type: int
        self.headers = None # type: str
        self.url_patch = None # type: str
        self.overwrite = None # type: bool
        self.raw_concat = None # type: bool
        self.disable_auto_concat = None # type: bool
        self.enable_auto_delete = None # type: bool
        self.disable_auto_decrypt = None # type: bool
        self.key = None # type: str
        self.b64key = None # type: str
        self.hexiv = None # type: str
        self.proxy = None # type: str
        self.disable_auto_exit = None # type: bool
        self.parse_only = None # type: bool
        self.show_init = None # type: bool
        self.index_to_name = None # type: bool
        self.log_level = None # type: str
        self.redl_code = None # type: list
        self.hide_load_metadata = None # type: bool
        self.no_metadata_file = None # type: bool
        self.gen_init_only = None # type: bool
        self.URI = None # type: list


import re
import math
from typing import List, Dict, Union
from .mpd import MPD
from .handler import xml_handler
from .childs.adaptationset import AdaptationSet
from .childs.role import Role
from .childs.baseurl import BaseURL
from .childs.contentprotection import ContentProtection
from .childs.period import Period
from .childs.representation import Representation
from .childs.s import S
from .childs.segmentlist import SegmentList
from .childs.initialization import Initialization
from .childs.segmenturl import SegmentURL
from .childs.segmentbase import SegmentBaee
from .childs.segmenttemplate import SegmentTemplate
from .childs.segmenttimeline import SegmentTimeline

from hlsdl.cmdargs import CmdArgs
from hlsdl.models.base import BaseUri
from hlsdl.extractors.base import BaseParser
from hlsdl.extractors.dash.key import DASHKey
from hlsdl.extractors.dash.stream import DASHStream
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class DASHParser(BaseParser):
    def __init__(self, args: CmdArgs, uri_type: str):
        super(DASHParser, self).__init__(args, uri_type)
        self.is_live = False
        self.root = None # type: MPD
        self.suffix = '.mpd'

    def fix_dash_base_url(self, previous_base_url: str, element: Union[MPD, Period, AdaptationSet, Representation]):
        '''
        BaseURL 可以是下面四个元素的子元素
        - MPD
        - Period
        - AdaptationSet
        - Representation

        根据多个渠道得到的BaseURL样本
        - 已有base_url路径上一级再拼接/all/ <BaseURL>../all/</BaseURL>
        - 已有base_url路径根路径拼接 <BaseURL>/12345/2145124</BaseURL>
        - 已有base_url路径常规拼接后续还需要拼接 <BaseURL>1280x720/</BaseURL>
        - 已有base_url路径常规拼接得到完整路径 <BaseURL>subtitles/TH/023719X0TH.vtt</BaseURL>
        - 已有base_url路径常规拼接得到完整路径 往往和SegmentBase搭配 <BaseURL>c0f671df-1314-4e9f-8aba-9ee42c26a69d_video_1.mp4</BaseURL>
        - 替换掉已有base_url 且通常有多个cdn <BaseURL serviceLocation="15802-Akamai" dvb:priority="3" dvb:weight="33">http://xxx.movetv.com/15802/live/PREM1/xxx/</BaseURL>

        每到一个层级都检查修正 修正后的base_url应当只适用于当前层级
        '''
        # logger.debug(f'previous_base_url {previous_base_url}')
        # 不管何时 previous_base_url 都应当以 http(s):// 开头
        if re.match(r'^https?://', previous_base_url) is None:
            assert False, 'unexcepted condition, report information to me'
        assert isinstance(element, (MPD, Period, AdaptationSet, Representation)), f'unexpected element type => {type(element)}'
        elements = element.find('BaseURL') # type: List[BaseURL]
        # 当前层级子元素中没有BaseURL元素
        # 那么原来的base_url是什么 那当前层级的base_url就是什么
        if len(elements) == 0:
            return previous_base_url
        base_url_element = None # type: BaseURL
        # 多个BaseURL则根据serviceLocation选择 否则选择第一个
        if len(elements) == 1:
            base_url_element = elements[0]
        else:
            # 没有指定 serviceLocation 则选第一个
            if not self.args.service:
                base_url_element = elements[0]
            else:
                # 如果BaseURL本身有serviceLocation 那么先尝试完全匹配
                for _element in elements:
                    if not _element.serviceLocation:
                        continue
                    if _element.serviceLocation == self.args.service:
                        base_url_element = _element
                        break
                if base_url_element is None:
                    # 不行再部分匹配
                    for _element in elements:
                        if not _element.serviceLocation:
                            continue
                        if self.args.service.lower() in _element.serviceLocation.lower():
                            base_url_element = _element
                            break
                if base_url_element is None:
                    # 还不行说明没有serviceLocation 那就选第一个
                    base_url_element = elements[0]
        # 先取内容
        base_url = base_url_element.innertext
        # 根据不同情况进行修正
        if re.match(r'^https?://', base_url):
            # 应当替换掉原有base_url
            return base_url
        if base_url.startswith('../'):
            # 对于这种情况应当去除末尾的/
            if previous_base_url.endswith('/'):
                previous_base_url = previous_base_url.rstrip('/')
            while base_url.startswith('../'):
                previous_base_url = '/'.join(previous_base_url.split("/")[:-1])
                base_url = base_url[3:]
            return previous_base_url + '/' + base_url
        if base_url.startswith('/'):
            items = previous_base_url.split('/')
            if len(items) >= 3:
                return '/'.join(items[:3]) + base_url
            else:
                # 最开始做过判断了 理论上不会走到这个分支 断言以防万一(...
                assert False, 'unexcepted condition, report information to me'
        if not previous_base_url.endswith('/'):
            previous_base_url += '/'
        return previous_base_url + base_url

    def parse(self, uri: str, content: str) -> List[DASHStream]:
        uri_item = self.parse_uri(uri)
        if uri_item is None:
            logger.error(f'parse {uri} failed')
            return []
        self.dump_content(uri_item.name, content, self.suffix)
        # 解析转换内容为期望的对象
        mpd = xml_handler(content)
        self.root = mpd
        # 检查是不是直播流
        if mpd.profiles == 'urn:mpeg:dash:profile:isoff-live:2011' and mpd.type != 'static':
            self.args.live = True
            self.is_live = True
        if self.args.live and self.is_live is False:
            logger.debug('detect current dash content is a living stream')
            self.is_live = True
        # 修正 MPD 节点的 BaseURL
        base_url = self.fix_dash_base_url(uri_item.base_url, mpd)
        return self.walk_period(mpd, uri_item.new_base_url(base_url))

    def walk_period(self, mpd: MPD, uri_item: BaseUri):
        periods = mpd.find('Period')  # type: List[Period]
        # 根据Period数量处理时长参数
        if len(periods) == 1 and periods[0].duration is None:
            # 当只存在一条流 且当前Period没有duration属性
            # 则使用mediaPresentationDuration作为当前Period的时长
            if hasattr(mpd, 'mediaPresentationDuration'):
                periods[0].duration = mpd.mediaPresentationDuration
            else:
                periods[0].duration = 0.0
        # 遍历处理periods
        streams = []
        for period in periods:
            # 修正 Period 节点的 BaseURL
            base_url = self.fix_dash_base_url(uri_item.base_url, period)
            _streams = self.walk_adaptationset(period, len(streams), uri_item.new_base_url(base_url))
            streams.extend(_streams)
        # 处理掉末尾的空分段
        for stream in streams:
            if stream.segments[-1].url == '':
                _ = stream.segments.pop(-1)
        # 合并流
        skey_stream = {} # type: Dict[str, DASHStream]
        for stream in streams:
            if stream.skey in skey_stream:
                skey_stream[stream.skey].update(stream, name_from_url=self.args.name_from_url)
            else:
                skey_stream[stream.skey] = stream
        streams = list(skey_stream.values())
        return streams

    def walk_adaptationset(self, period: Period, sindex: int, uri_item: BaseUri):
        adaptationsets = period.find('AdaptationSet') # type: List[AdaptationSet]
        streams = []
        for adaptationset in adaptationsets:
            # 修正 AdaptationSet 节点的 BaseURL
            base_url = self.fix_dash_base_url(uri_item.base_url, adaptationset)
            current_uri_item = uri_item.new_base_url(base_url)
            if adaptationset.mimeType == 'image/jpeg':
                logger.debug(f'skip parse for AdaptationSet mimeType image/jpeg')
                continue
            representations = adaptationset.find('Representation') # type: List[Representation]
            if len(representations) > 0:
                _streams = self.walk_representation(adaptationset, period, sindex + len(streams), current_uri_item)
            else:
                assert False, 'not implemented yet, report this mpd content to me'
                segmenttemplates = adaptationset.find('SegmentTemplate') # type: List[SegmentTemplate]
                assert len(segmenttemplates) == 1, 'plz report this mpd content to me'
                segmenttimelines = segmenttemplates[0].find('SegmentTimeline') # type: List[SegmentTimeline]
                _streams = self.walk_s_v2(segmenttimelines[0], adaptationset, period, sindex + len(streams), current_uri_item)
            streams.extend(_streams)
        return streams

    def walk_representation(self, adaptationset: AdaptationSet, period: Period, sindex: int, uri_item: BaseUri):
        '''
        每一个<Representation></Representation>都对应轨道的一/整段
        '''
        representations = adaptationset.find('Representation') # type: List[Representation]
        segmenttemplates = adaptationset.find('SegmentTemplate') # type: List[SegmentTemplate]
        streams = []
        for representation in representations:
            # 修正 Representation 节点的 BaseURL
            base_url = self.fix_dash_base_url(uri_item.base_url, representation)
            current_uri_item = uri_item.new_base_url(base_url)
            if self.args.log_level == 'DEBUG':
                logger.debug(f'current_base_url {current_uri_item.base_url}')
            stream = DASHStream(sindex, current_uri_item, self.args.save_dir)
            sindex += 1
            self.walk_contentprotection(adaptationset, stream)
            self.walk_contentprotection(representation, stream)
            # 给流设置属性
            stream.set_skey(adaptationset.id, representation.id)
            stream.set_lang(adaptationset.lang)
            stream.set_bandwidth(representation.bandwidth)
            if representation.codecs is None:
                stream.set_codecs(adaptationset.codecs)
            else:
                stream.set_codecs(representation.codecs)
            if representation.mimeType is None:
                stream.set_stream_type(adaptationset.mimeType)
            else:
                stream.set_stream_type(representation.mimeType)
            if representation.width is None or representation.height is None:
                stream.set_resolution(adaptationset.width, adaptationset.height)
            else:
                stream.set_resolution(representation.width, representation.height)
            # 针对字幕直链类型
            Roles = adaptationset.find('Role') # type: List[Role]
            if stream.stream_type == '' and len(Roles) > 0:
                stream.set_stream_type(Roles[0].value)
            segmentlists = representation.find('SegmentList') # type: List[SegmentList]
            r_segmenttemplates = representation.find('SegmentTemplate') # type: List[SegmentTemplate]
            # 针对视频音频流处理 分情况生成链接
            if len(segmentlists) == 1:
                self.walk_segmentlist(segmentlists[0], representation, period, stream)
            elif len(segmenttemplates) == 0:
                self.walk_segmenttemplate(representation, period, stream)
            elif len(segmenttemplates) == 1 and len(segmenttemplates[0].find('SegmentTimeline')) == 1:
                self.walk_segmenttimeline(segmenttemplates[0], representation, period, stream)
            elif len(r_segmenttemplates) == 1 and len(r_segmenttemplates[0].find('SegmentTimeline')) == 1:
                self.walk_segmenttimeline(r_segmenttemplates[0], representation, period, stream)
            elif len(segmenttemplates) == 1 and segmenttemplates[0].initialization is None:
                # tv-player.ap1.admint.biz live
                _segmenttemplates = representation.find('SegmentTemplate')
                if len(_segmenttemplates) != 1:
                    # AdaptationSet 的 SegmentTemplate 没有 initialization 
                    # Representation 没有 SegmentTemplate 则跳过
                    continue
                # assert len(_segmenttemplates) == 1, '请报告出现此异常提示的mpd/report plz'
                segmenttemplate = segmenttemplates[0]
                _segmenttemplate = _segmenttemplates[0]
                if segmenttemplate.timescale is not None:
                    _segmenttemplate.timescale = segmenttemplate.timescale
                if segmenttemplate.duration is not None:
                    _segmenttemplate.duration = segmenttemplate.timescale
                self.generate_v1(period, representation.id, _segmenttemplate, stream)
            else:
                # SegmentTemplate 和多个 Representation 在同一级
                # 那么 SegmentTemplate 的时长参数等就是多个 Representation 的参数
                # 同一级的时候 只有一个 SegmentTemplate
                self.generate_v1(period, representation.id, segmenttemplates[0], stream)
            streams.append(stream)
        return streams

    def walk_segmentlist(self, segmentlist: SegmentList, representation: Representation, period: Period, stream: DASHStream):
        initializations = segmentlist.find('Initialization') # type: List[Initialization]
        has_initialization = False
        if len(initializations) == 1:
            has_initialization = True
            stream.set_init_url(initializations[0].sourceURL)
        segmenturls = segmentlist.find('SegmentURL') # type: List[SegmentURL]
        for segmenturl in segmenturls:
            if segmenturl.media == '':
                # 通常就是一个整段
                continue
            stream.set_media_url(segmenturl.media, name_from_url=self.args.name_from_url)
        if has_initialization:
            interval = float(segmentlist.duration / segmentlist.timescale)
            stream.set_segments_duration(interval)

    def walk_contentprotection(self, representation: Representation, stream: DASHStream):
        ''' 流的加密方案 '''
        contentprotections = representation.find('ContentProtection') # type: List[ContentProtection]
        for contentprotection in contentprotections:
            # DASH流的解密通常是合并完整后一次解密
            # 不适宜每个分段单独解密
            # 那么这里就不用给每个分段设置解密key了
            # 而且往往key不好拿到 所以这里仅仅做一个存储
            stream.append_key(DASHKey(contentprotection))

    def walk_segmenttemplate(self, representation: Representation, period: Period, stream: DASHStream):
        baseurls = representation.find('BaseURL') # type: List[BaseURL]
        segmentbases = representation.find('SegmentBase') # type: List[SegmentBaee]
        if len(segmentbases) == 1 and len(baseurls) == 1:
            if stream.base_url.startswith('http') or stream.base_url.startswith('/'):
                stream.set_init_url(stream.base_url)
            else:
                # set baseurls[0].innertext.strip() ?
                stream.set_init_url('../' + stream.base_url)
            # stream.set_segment_duration(-1)
            return
        segmenttemplates = representation.find('SegmentTemplate') # type: List[SegmentTemplate]
        # segmenttimelines = representation.find('SegmentTimeline') # type: List[SegmentTimeline]
        if len(segmenttemplates) != 1:
            # 正常情况下 这里应该只有一个SegmentTemplate
            # 没有就无法计算分段 则跳过
            # 不止一个可能是没见过的类型 提醒上报
            if len(segmenttemplates) > 1:
                logger.error('please report this DASH content.')
            else:
                # logger.warning('stream has no SegmentTemplate between Representation tag.')
                if stream.base_url.startswith('http'):
                    stream.set_init_url(stream.base_url)
            return
        if len(segmenttemplates[0].find('SegmentTimeline')) == 0:
            self.generate_v1(period, representation.id, segmenttemplates[0], stream)
            return
        self.walk_segmenttimeline(segmenttemplates[0], representation, period, stream)

    def walk_segmenttimeline(self, segmenttemplate: SegmentTemplate, representation: Representation, period: Period, stream: DASHStream):
        segmenttimelines = segmenttemplate.find('SegmentTimeline') # type: List[SegmentTimeline]
        if len(segmenttimelines) != 1:
            if len(segmenttimelines) > 1:
                logger.error('please report this DASH content.')
            else:
                logger.warning('stream has no SegmentTimeline between SegmentTemplate tag.')
            return
        self.walk_s(segmenttimelines[0], segmenttemplate, representation, period, stream)

    def walk_s_v2(self, segmenttimeline: SegmentTimeline, adaptationset: AdaptationSet, period: Period, sindex: int, uri_item: BaseUri):
        stream = DASHStream(sindex, uri_item, self.args.save_dir)
        stream.set_skey(adaptationset.id, None)
        sindex += 1
        return [stream]

    def walk_s(self, segmenttimeline: SegmentTimeline, st: SegmentTemplate, representation: Representation, period: Period, stream: DASHStream):
        init_url = st.get_url()
        if init_url is not None:
            if '$RepresentationID$' in init_url:
                init_url = init_url.replace('$RepresentationID$', representation.id)
            if '$Bandwidth$' in init_url:
                init_url = init_url.replace('$Bandwidth$', str(representation.bandwidth))
            if re.match('.*?as=audio_(.*?)\)', init_url):
                _lang = re.match('.*?as=audio_(.*?)\)', init_url).groups()[0]
                stream.set_lang(_lang)
            stream.set_init_url(init_url)
        else:
            # 这种情况可能是因为流是字幕
            pass
        target_r = 0 # type: int
        ss = segmenttimeline.find('S') # type: List[S]
        if len(ss) > 0 and self.is_live and ss[0].t > 0:
            # 这个部分是修正 base_time
            # timeShiftBufferDepth => cdn max cache time for segments
            # newest available segment $Time$ should meet below condition
            # SegmentTimeline.S.t / timescale + (mpd.availabilityStartTime + Period.start) <= time.time()
            base_time = None # type: int
            assert isinstance(self.root.availabilityStartTime, float), 'report mpd to me'
            current_utctime = self.root.publishTime.timestamp() - self.args.live_utc_offset
            presentation_start = period.start - st.presentationTimeOffset / st.timescale + 30
            start_utctime = self.root.availabilityStartTime + presentation_start
            if self.args.log_level == 'DEBUG':
                logger.debug(f'mpd.presentationTimeOffset {st.presentationTimeOffset} timescale {st.timescale}')
                logger.debug(f'mpd.availabilityStartTime {self.root.availabilityStartTime} Period.start {period.start}')
                logger.debug(f'start_utctime {start_utctime} current_utctime {current_utctime}')
            tmp_t = ss[0].t
            for s in ss:
                for number in range(s.r):
                    if (tmp_t + s.d) / st.timescale + start_utctime > current_utctime:
                        base_time = tmp_t
                        if self.args.log_level == 'DEBUG':
                            logger.debug(f'set base_time {base_time} target_r {target_r}')
                        break
                    if target_r > 0:
                        tmp_t += s.d
                    target_r += 1
                if base_time:
                    break
            if base_time is None:
                logger.debug(f'{representation.id} report mpd to me, maybe need wait {current_utctime - start_utctime - tmp_t / st.timescale}s')
            assert base_time is not None, f'{representation.id} report mpd to me, maybe need wait {current_utctime - start_utctime - tmp_t / st.timescale}s'
            # if base_time is None:
            #     base_time = ss[0].t
        elif ss[0].t > 0:
            base_time = ss[0].t
            logger.debug(f'ss[0].t > 0, set base_time {base_time}')
        else:
            base_time = 0
        # 如果 base_time 不为 0 即第一个 s.t 不为
        # 那么 time_offset 就不需要 即设置为 0
        time_offset = st.presentationTimeOffset if base_time == 0 else 0
        start_number = st.startNumber
        tmp_offset_r = 0
        total_segments_duration = 0.0
        for index, s in enumerate(ss):
            if self.args.multi_s and index > 0 and s.t > 0:
                base_time = s.t
            if st.timescale == 0:
                interval = 0
            else:
                interval = s.d / st.timescale
            if s.r == -1:
                _range = math.ceil((period.duration or self.root.mediaPresentationDuration) / interval)
            else:
                _range = s.r
            for number in range(_range):
                tmp_offset_r += 1
                if self.is_live and tmp_offset_r < target_r:
                    continue
                # 经过测试 对于直播流 应当计算时长来确定应该下载的分段
                # 因为无法通过比较两轮的url来排除已经下载的分段 通过url比较会导致重复下载
                # 根据标准 这里与 minBufferTime 或者 minimumUpdatePeriod 比较都可以 浏览器是后者 保持一致
                if self.is_live and total_segments_duration > self.root.minimumUpdatePeriod:
                    break
                total_segments_duration += interval
                media_url = st.get_media_url()
                if '$Bandwidth$' in media_url:
                    media_url = media_url.replace('$Bandwidth$', str(representation.bandwidth))
                if '$Number$' in media_url:
                    media_url = media_url.replace('$Number$', str(start_number))
                    start_number += 1
                if re.match('.*?(\$Number%(.+?)\$)', media_url):
                    old, fmt = re.match('.*?(\$Number(%.+?)\$)', media_url).groups()
                    new = fmt % start_number
                    media_url = media_url.replace(old, new)
                    start_number += 1
                if '$RepresentationID$' in media_url:
                    media_url = media_url.replace('$RepresentationID$', representation.id)
                if '$Time$' in media_url:
                    fmt_time = time_offset + base_time
                    stream.set_segment_fmt_time(fmt_time)
                    media_url = media_url.replace('$Time$', str(fmt_time))
                    time_offset += s.d
                stream.set_segment_duration(interval)
                stream.set_media_url(media_url, name_from_url=self.args.name_from_url)

    def generate_v1(self, period: Period, rid: str, st: SegmentTemplate, stream: DASHStream):
        init_url = st.get_url()
        if '$RepresentationID$' in init_url:
            init_url = init_url.replace('$RepresentationID$', rid)
        stream.set_init_url(init_url)
        if st.timescale == 0:
            interval = st.duration
        else:
            interval = float(int(st.duration) / int(st.timescale))
        repeat = math.ceil(period.duration / interval)
        for number in range(int(st.startNumber), repeat + int(st.startNumber)):
            media_url = st.get_media_url()
            if '$Number$' in media_url:
                media_url = media_url.replace('$Number$', str(number))
            if re.match('.*?(\$Number%(.+?)\$)', media_url):
                old, fmt = re.match('.*?(\$Number(%.+?)\$)', media_url).groups()
                new = fmt % number
                media_url = media_url.replace(old, new)
            if '$RepresentationID$' in media_url:
                media_url = media_url.replace('$RepresentationID$', rid)
            stream.set_media_url(media_url, name_from_url=self.args.name_from_url)
        stream.set_segments_duration(interval)


def tree(obj, step: int = 0):
    print(f"{step * '--'}>{obj.name}")
    step += 1
    for child in obj.childs:
        step = tree(child, step=step)
    step -= 1
    print(f"{step * '--'}>{obj.name}")
    return step


from pathlib import Path
from hlsdl.cmdargs import CmdArgs
from hlsdl.models.base import BaseUri
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class BaseParser:
    def __init__(self, args: CmdArgs, uri_type: str):
        self.args = args
        self.uri_type = uri_type
        self.suffix = '.SUFFIX'

    def fix_name(self, name: str):
        '''
        remove illegal char
        '''
        # logger.debug(f'fix name before: {name}')
        exclude_str = ["\\", "/", ":", "：", "*", "?", "\"", "<", ">", "|", "\r", "\n", "\t"]
        for s in exclude_str:
            name = name.replace(s, " ")
        name = "_".join(name.split())
        # logger.debug(f'fix name after: {name}')
        return name

    def dump_content(self, name: str, content: str, suffix: str):
        if self.args.no_metadata_file:
            return
        dump_path = self.args.save_dir / f'{name}{suffix}'
        logger.debug(f'save content to {dump_path.resolve().as_posix()}, size {len(content)}')
        dump_path.write_text(content, encoding='utf-8')

    def parse_uri(self, uri: str) -> BaseUri:
        '''
        进入此处的uri不可能是文件夹
        '''
        logger.debug(f'start parse uri for: {uri}')
        rm_manifest = False
        if '.ism' in self.args.base_url and 'manifest' in self.args.base_url:
            rm_manifest = True
        if '.ism' in uri and 'manifest' in uri:
            rm_manifest = True
        name = self.args.name
        if self.uri_type == 'path':
            name = Path(uri).stem
        home_url, base_url = '', ''
        if uri.startswith('http://') or uri.startswith('https://') or uri.startswith('ftp://'):
            uris = uri.split('?', maxsplit=1)
            if name == '':
                name = uris[0][::-1].split('/', maxsplit=1)[0][::-1]
            if name.endswith(self.suffix):
                name = name[:-len(self.suffix)]
            home_url = '/'.join(uris[0].split('/', maxsplit=3)[:-1])
            base_url = uris[0][::-1].split('/', maxsplit=1)[-1][::-1]
        elif Path(uri).exists():
            if name == '':
                name = Path(uri).stem
        if base_url == '' and self.args.base_url != '':
            if rm_manifest and self.args.base_url.rstrip('/').endswith('/manifest'):
                base_url = '/'.join(self.args.base_url.rstrip('/').split('/')[:-1])
            else:
                base_url = self.args.base_url
            home_url = '/'.join(base_url.split('/', maxsplit=3)[:-1])
        name = self.fix_name(name)
        logger.debug(
            f'parse uri result:\n'
            f'    name {name}\n'
            f'    home_url {home_url}\n'
            f'    base_url {base_url}'
        )
        # return [name, home_url, base_url]
        return BaseUri(name, home_url, base_url)


from .x import X


class XPrivinf(X):
    '''
    #EXT-X-PRIVINF YOUKU的自定义标签
    - FILESIZE=925655
    '''
    def __init__(self):
        super(XPrivinf, self).__init__('#EXT-X-PRIVINF')
        self.filesize = None # type: int
        self.drm_notencrypt = False # type: bool
        self.known_attrs = {
            'FILESIZE': int,
        }

    def set_attrs_from_line(self, line: str):
        if line.endswith('DRM_NOTENCRYPT'):
            self.drm_notencrypt = True
            line = line.replace('DRM_NOTENCRYPT', '')
        return super(XPrivinf, self).set_attrs_from_line(line)


import sys
import base64
import logging
import platform
from pathlib import Path
from argparse import ArgumentParser

from hlsdl.cmdargs import CmdArgs
from hlsdl.daemon import Daemon
from hlsdl.version import __version__
from hlsdl.headers.default import Headers
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


def command_handler(args: CmdArgs):
    '''
    对命令参数进行校验和修正
    '''
    args.speed_up_left = int(args.speed_up_left)
    if args.live_duration == '':
        args.live_duration = 0.0
    else:
        hms = args.live_duration.split(':')
        assert len(hms) == 3, '--live-duration format is HH:MM:SS, example: 00:00:30'
        assert len(hms[1]) <= 2, '--live-duration minute length must less than or equal to 2'
        assert len(hms[2]) <= 2, '--live-duration second length must less than or equal to 2'
        if hms[0].isdigit() and hms[1].isdigit() and hms[2].isdigit():
            assert float(hms[1]) <= 60.0, '--live-duration minute must less than or equal to 60'
            assert float(hms[2]) <= 60.0, '--live-duration second must less than or equal to 60'
            args.live_duration = float(hms[0]) * 60 * 60 + float(hms[1]) * 60 + float(hms[2])
    try:
        args.live_utc_offset = int(args.live_utc_offset)
        args.live_refresh_interval = int(args.live_refresh_interval)
    except Exception:
        assert False, '--live-utc-offset can not be convert to int value'
    if args.video_only is True and args.audio_only is True:
        assert False, '--video-only and --audio-only cannot be used at the same time'
    args.save_dir = Path(args.save_dir)
    if args.save_dir.exists() is False:
        args.save_dir.mkdir()
    args.headers = Headers().get(args)
    args.limit_per_host = int(args.limit_per_host)
    if args.key is not None:
        infos = args.key.split(':')
        assert len(infos) == 2, 'DASH Stream decryption key format error !'
        # assert len(infos[0]) == 32, 'DASH Stream decryption key @KID must be 32 length hex string !'
        assert len(infos[1]) == 32, 'DASH Stream decryption key @k must be 32 length hex string !'
    if args.b64key is not None:
        try:
            _ = base64.b64decode(args.b64key)
        except Exception as e:
            raise e
    if args.hexiv is not None:
        if args.hexiv.lower().startswith('0x'):
            args.hexiv = args.hexiv.lower()[2:]

    if getattr(sys, 'frozen', False):
        bin_path = Path(sys.executable).parent / 'binaries'
    else:
        bin_path = Path(__file__).parent.parent / 'binaries'
    if bin_path.exists() is False:
        args.ffmpeg = 'ffmpeg'
        args.mp4decrypt = 'mp4decrypt'
        args.mp4box = 'mp4box'
        logger.warning(f'binaries folder is not exist > {bin_path}')
    else:
        if platform.system() == 'Windows':
            args.ffmpeg = (bin_path / 'ffmpeg.exe').resolve().as_posix()
            args.mp4decrypt = (bin_path / 'mp4decrypt.exe').resolve().as_posix()
            args.mp4box = (bin_path / 'mp4box.exe').resolve().as_posix()
        else:
            args.ffmpeg = (bin_path / 'ffmpeg').resolve().as_posix()
            args.mp4decrypt = (bin_path / 'mp4decrypt').resolve().as_posix()
            args.mp4box = (bin_path / 'mp4box').resolve().as_posix()
    logger.debug(f'ffmpeg {args.ffmpeg}')
    logger.debug(f'mp4decrypt {args.mp4decrypt}')
    logger.debug(f'mp4box {args.mp4box}')
    try:
        args.redl_code = [int(_.strip()) for _ in args.redl_code.split(',') if _ != '']
    except Exception as e:
        logger.error(f'parse --redl-code option failed', exc_info=e)
        args.redl_code = []


def main():
    def print_version():
        print(f'version {__version__}, A downloader that download the HLS/DASH stream.')

    parser = ArgumentParser(prog='XstreamDL-CLI', usage='XstreamDL-CLI [OPTION]... URL/FILE/FOLDER...', description='A downloader that download the HLS/DASH stream', add_help=False)
    parser.add_argument('-v', '--version', action='store_true', help='print version and exit')
    parser.add_argument('-h', '--help', action='store_true', help='print help message and exit')
    parser.add_argument('--speed-up', action='store_true', help='speed up at end')
    parser.add_argument('--speed-up-left', default='10', help='speed up when left count less than this value')
    parser.add_argument('--live', action='store_true', help='live mode')
    parser.add_argument('--compare-with-url', action='store_true', help='use full url to compare with last segments to get new segments')
    parser.add_argument('--dont-split-discontinuity', action='store_true', help='dont take #EXT-X-DISCONTINUITY tag as a new stream')
    parser.add_argument('--name-from-url', action='store_true', help='get name from segment url')
    parser.add_argument('--live-duration', default='', help='live record time, format HH:MM:SS, example 00:00:30 will record about 30s')
    parser.add_argument('--live-utc-offset', default='0', help='the value is used to correct utc time')
    parser.add_argument('--live-refresh-interval', default='3', help='live refresh interval')
    parser.add_argument('--name', default='', help='specific stream base name')
    parser.add_argument('--base-url', default='', help='set base url for Stream')
    parser.add_argument('--ad-keyword', default='', help='skip #EXT-X-DISCONTINUITY which segment url has this keyword')
    parser.add_argument('--resolution', default='', choices=['', '270', '360', '480', '540', '576', '720', '1080', '2160'], help='auto choose target quality')
    parser.add_argument('--best-quality', action='store_true', help='auto choose best quality for dash streams')
    parser.add_argument('--video-only', action='store_true', help='only choose video stream when use --best-quality')
    parser.add_argument('--audio-only', action='store_true', help='only choose audio stream when use --best-quality')
    parser.add_argument('--all-videos', action='store_true', help='choose all video stream to download')
    parser.add_argument('--all-audios', action='store_true', help='choose all audio stream to download')
    parser.add_argument('--all-subtitles', action='store_true', help='choose all subtitle stream to download')
    parser.add_argument('--service', default='', help='set serviceLocation for BaseURL choose')
    parser.add_argument('--save-dir', default='Downloads', help='set save dir for Stream')
    parser.add_argument('--select', action='store_true', help='show stream to select and download, default is to download all')
    parser.add_argument('--multi-s', action='store_true', help='use this option when S tag number > 0')
    parser.add_argument('--disable-force-close', action='store_true', help='default make all connections closed securely, but it will make DL speed slower')
    parser.add_argument('--limit-per-host', default=4, help='increase the value if your connection to the stream host is poor, suggest >100 for DASH stream')
    parser.add_argument('--headers', default='headers.json', help='read headers from headers.json, you can also use custom config')
    parser.add_argument('--url-patch', default='', help='add some custom strings for all segments link')
    parser.add_argument('--overwrite', action='store_true', help='overwrite output files')
    parser.add_argument('--raw-concat', action='store_true', help='concat content as raw')
    parser.add_argument('--disable-auto-concat', action='store_true', help='disable auto-concat')
    parser.add_argument('--enable-auto-delete', action='store_true', help='enable auto-delete files after concat success')
    parser.add_argument('--disable-auto-decrypt', action='store_true', help='disable auto-decrypt segments before dump to disk')
    parser.add_argument('--key', default=None, help='<id>:<k>, <id> is either a track ID in decimal or a 128-bit KID in hex, <k> is a 128-bit key in hex')
    parser.add_argument('--b64key', default=None, help='base64 format aes key, only for HLS standard AES-128-CBC encryption')
    parser.add_argument('--hexiv', default=None, help='hex format aes iv')
    parser.add_argument('--proxy', default='', help='use socks/http proxy, e.g. socks5://127.0.0.1:10808 or http://127.0.0.1:10809')
    parser.add_argument('--disable-auto-exit', action='store_true', help='disable auto exit after download end, GUI will use this option')
    parser.add_argument('--parse-only', action='store_true', help='parse only, not to download')
    parser.add_argument('--show-init', action='store_true', help='show initialization to help you identify same name stream')
    parser.add_argument('--index-to-name', action='store_true', help='some dash live have the same name for different stream, use this option to avoid')
    parser.add_argument('--log-level', default='INFO', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'], help='set log level, default is INFO')
    parser.add_argument('--redl-code', default='502', help='re-download set of response status codes , e.g. 408,500,502,503,504')
    parser.add_argument('--hide-load-metadata', action='store_true', help='hide `Load #EXT-X-MEDIA metadata` balabala')
    parser.add_argument('--no-metadata-file', action='store_true', help='do not save metadata file(.m3u8/.mpd/.ism/...)')
    parser.add_argument('--gen-init-only', action='store_true', help='generate init segment only')
    parser.add_argument('URI', nargs='*', help='URL/FILE/FOLDER string')
    args = parser.parse_args()
    if args.help:
        print_version()
        parser.print_help()
        sys.exit()
    if args.version:
        print_version()
        sys.exit()
    if len(args.URI) == 0:
        try:
            uri = input('Paste your URL/FILE/FOLDER string at the end of commands, plz.\nCtrl C to exit or input here:')
        except KeyboardInterrupt:
            sys.exit()
        if uri.strip() != '':
            args.URI.append(uri.strip())
    if len(args.URI) == 0:
        sys.exit('No URL/FILE/FOLDER input')
    for handler in logger.handlers:
        if isinstance(handler, logging.FileHandler) and args.log_level != 'DEBUG':
            logger.handlers.remove(handler)
            break
    for handler in logger.handlers:
        # 注意 这里不能拿 StreamHandler 做判断
        # 因为 FileHandler 父类是 StreamHandler
        # 这样当 handler 是 FileHandler 的时候 isinstance 返回 True
        if isinstance(handler, logging.FileHandler) is False:
            handler.setLevel(logging.getLevelName(args.log_level))
    command_handler(args)
    logger.info(f'use {__version__}, set URI to {args.URI}')
    logger.debug(f'args => {args}')
    daemon = Daemon(args)
    daemon.daemon()
    if args.disable_auto_exit:
        _ = input('press any key to exit.')


if __name__ == '__main__':
    main()


import sys
import math
import time
import signal
import asyncio
import binascii
from typing import List, Set, Dict
from asyncio import new_event_loop
from asyncio import AbstractEventLoop, Future, Task
from aiohttp import client_exceptions
from aiohttp import ClientResponse, ClientSession, ClientTimeout, TCPConnector
from aiohttp_socks import ProxyConnector
from concurrent.futures._base import TimeoutError, CancelledError
from hlsdl.cmdargs import CmdArgs
from hlsdl.models.stream import Stream
from hlsdl.models.segment import Segment
from hlsdl.util.decryptors.aes import CommonAES
from hlsdl.util.texts import t_msg
from hlsdl.log import setup_logger

logger = setup_logger('hlsdl', level='INFO')


def auto_choose_resolution(args: CmdArgs, streams: List[Stream]) -> List[Stream]:
    target_indexes = []
    for index, stream in enumerate(streams):
        # 跳过非目标类型
        if stream.stream_type != 'video':
            continue
        # 如果当前流的目标分辨率与预设分辨率不匹配 跳过
        if stream.resolution.split('x')[1] != args.resolution:
            continue
        target_indexes.append(index)
    return target_indexes


def auto_choose_best_streams(args: CmdArgs, streams: List[Stream]) -> List[Stream]:
    def choose_best_stream_by_type(target_type: str):
        ''' return target index '''
        target_stream = None
        for stream in streams:
            # 跳过非目标类型
            if stream.stream_type != target_type:
                continue
            # 跳过没有bandwidth的流
            if not stream.bandwidth:
                continue
            # 目标是视频并且下载选项设定了目标分辨率
            if target_type == 'video' and args.resolution != '':
                # 如果当前流的目标分辨率与预设分辨率不匹配 跳过
                if stream.resolution.split('x')[1] != args.resolution:
                    continue
            # 如果target_stream还没设置 当前流设置为target_stream
            if target_stream is None:
                target_stream = stream
                continue
            # 当前流码率比之前选定的流更高 设置当前流为选择的流
            if stream.bandwidth > target_stream.bandwidth:
                target_stream = stream
        # 在上面的检查中确实设置了流 则返回其索引
        # 这里其实返回流更合理 但是上一级其他位置一直用的索引 所以这里先不改
        if target_stream:
            return streams.index(target_stream)
        # 否则返回None 交由上一级判断
        return target_stream
    best_audio_stream_index = choose_best_stream_by_type('audio')
    best_video_stream_index = choose_best_stream_by_type('video')
    if args.audio_only:
        if best_audio_stream_index is not None:
            return [best_audio_stream_index]
        else:
            return []
    if args.video_only:
        if best_video_stream_index is not None:
            return [best_video_stream_index]
        else:
            return []
    target_streams = []
    if best_audio_stream_index is not None:
        target_streams.append(best_audio_stream_index)
    if best_video_stream_index is not None:
        target_streams.append(best_video_stream_index)
    return target_streams


def get_selected_index(length: int) -> list:
    selected = []
    try:
        text = input(t_msg.input_stream_number).strip()
    except EOFError:
        return []
    if text == '':
        return [index for index in range(length + 1)]
    elif text.isdigit():
        return [int(text)]
    elif '-' in text and len(text.split('-')) == 2:
        start, end = text.split('-')
        if start.strip().isdigit() and end.strip().isdigit():
            return [index for index in range(int(start.strip()), int(end.strip()) + 1)]
    elif text.replace(' ', '').isdigit():
        for index in text.split(' '):
            if index.strip().isdigit():
                if int(index.strip()) <= length:
                    selected.append(int(index))
        return selected
    elif text.replace(',', '').replace(' ', '').isdigit():
        for index in text.split(','):
            if index.strip().isdigit():
                if int(index.strip()) <= length:
                    selected.append(int(index))
        return selected
    return selected


def get_left_segments(stream: Stream):
    count = 0
    completed = 0
    _left_segments = []
    for segment in stream.segments:
        segment_path = segment.get_path()
        if segment_path.exists() is True:
            # 文件落盘 说明下载一定成功了
            if segment_path.stat().st_size == 0:
                segment_path.unlink()
            else:
                count += 1
                completed += segment_path.stat().st_size
                continue
        _left_segments.append(segment)
    return count, completed, _left_segments


def get_connector(args: CmdArgs):
    '''
    connector在一个ClientSession使用后可能就会关闭
    若需要再次使用则需要重新生成
    '''
    if args.proxy != '':
        return ProxyConnector.from_url(
            args.proxy,
            ttl_dns_cache=500,
            ssl=False,
            limit_per_host=args.limit_per_host,
            limit=500,
            force_close=not args.disable_force_close,
            enable_cleanup_closed=not args.disable_force_close
        )
    return TCPConnector(
        ttl_dns_cache=500,
        ssl=False,
        limit_per_host=args.limit_per_host,
        limit=500,
        force_close=not args.disable_force_close,
        enable_cleanup_closed=not args.disable_force_close
    )


class XProgress:

    def __init__(self, title: str, total_count: int, downloaded_count: int, total_size: int, completed_size: int, speed_up_flag: bool, speed_up_left: int):
        self.last_time = time.time()
        self.title = title
        self.total_count = total_count
        self.downloaded_count = downloaded_count
        self.total_size = total_size
        self.downloaded_size = completed_size
        self.last_size = completed_size
        self.speed_up_flag = speed_up_flag
        self.speed_up_left = speed_up_left
        self.stop = False

    def is_ending(self):
        if self.speed_up_flag is False:
            return False
        return self.total_count - self.downloaded_count < self.speed_up_left

    def calc_speed(self, total_size: int, downloaded_size: int):
        ts = time.time()
        tm = ts - self.last_time
        if self.stop is False and tm < 0.3:
            return
        if tm == 0.0:
            return
        speed = (downloaded_size - self.last_size) / tm / 1024 / 1024
        self.last_time = ts
        self.last_size = downloaded_size
        return speed

    def add_downloaded_count(self, downloaded_count: int):
        self.downloaded_count += downloaded_count
        self.update_progress(self.downloaded_count, self.total_size, self.downloaded_size)

    def update_total_size(self, total_size: int):
        self.total_size = total_size
        self.update_progress(self.downloaded_count, self.total_size, self.downloaded_size)

    def decrease_total_count(self):
        self.total_count -= 1
        self.update_progress(self.downloaded_count, self.total_size, self.downloaded_size)

    def add_downloaded_size(self, downloaded_size: int):
        self.downloaded_size += downloaded_size
        self.update_progress(self.downloaded_count, self.total_size, self.downloaded_size)

    def update_progress(self, downloaded_count: int, total_size: int, downloaded_size: int):
        barlen, status = 30, ''
        progress = downloaded_count / self.total_count
        if progress >= 1.0:
            progress, status = 1, '\r\n'
        speed = self.calc_speed(total_size, downloaded_size)
        if speed is None:
            return
        _total_size = total_size / 1024 / 1024
        _downloaded_size = downloaded_size / 1024 / 1024
        bar_str = chr(9608)
        split_str = chr(8226)
        block = int(math.floor(barlen * progress))
        bar = bar_str * block + ' ' * (barlen - block)
        text = (
            f'\r{self.title} {bar} {_downloaded_size:.2f}/{_total_size:.2f}MB {split_str} {speed:.2f}MB/s '
            f'{split_str} {downloaded_count}/{self.total_count} {split_str} {progress * 100:.2f}% {status}'
        )
        sys.stdout.write(text)
        sys.stdout.flush()

    def to_stop(self, is_error: bool = False):
        self.stop = True
        self.update_progress(self.downloaded_count, self.total_size, self.downloaded_size)
        if is_error:
            sys.stdout.write('\r\n')
            sys.stdout.flush()


class Downloader:

    def __init__(self, args: CmdArgs):
        self.args = args
        self.log_detail = args.log_level.upper() == 'DEBUG'
        self.xprogress = None # type: XProgress
        self.terminate = False
        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)

    def stop(self, signum: int, frame):
        logger.debug('stopped by Ctrl C')
        self.terminate = True

    def stop_record(self):
        logger.debug('stopped reason: stop_record')
        self.terminate = True

    def do_select(self, streams: List[Stream], selected: list = []):
        if len(selected) > 0:
            return selected
        if streams is None:
            return
        if len(streams) == 0:
            return
        for index, stream in enumerate(streams):
            stream.show_info(index, show_init=self.args.show_init, index_to_name=self.args.index_to_name)
        if self.args.select is True:
            selected = get_selected_index(len(streams))
            if len(selected) == 0:
                logger.info(t_msg.select_without_any_stream)
        elif self.args.best_quality:
            selected = auto_choose_best_streams(self.args, streams)
        elif self.args.all_videos:
            selected = [index for index, stream in enumerate(streams) if stream.stream_type == 'video']
        elif self.args.all_audios:
            selected = [index for index, stream in enumerate(streams) if stream.stream_type == 'audio']
        elif self.args.all_subtitles:
            selected = [index for index, stream in enumerate(streams) if stream.stream_type == 'subtitle']
        elif self.args.resolution != '':
            selected = auto_choose_resolution(self.args, streams)
        else:
            selected = [index for index in range(len(streams))]
        if self.args.live is False:
            return selected
        skeys = []
        for select in selected:
            skeys.append(streams[select].get_skey())
        return skeys

    def download_streams(self, streams: List[Stream], selected: list = []):
        selected = self.do_select(streams, selected)
        logger.debug(f'selected is {selected}')
        if selected is None:
            return
        should_stop_record = False
        all_results = []
        for index, stream in enumerate(streams):
            if self.terminate is True:
                break
            if self.args.live is False:
                if index not in selected:
                    continue
            else:
                if stream.get_skey() not in selected:
                    continue
            if self.args.live is False and len(stream.segments) == 1:
                logger.warning(f'only one segment, download speed maybe slow =>\n{stream.segments[0].url + self.args.url_patch}')
                # continue
            stream.dump_segments()
            max_failed = 5
            if self.args.parse_only:
                if len(stream.segments) <= 5:
                    stream.show_segments()
                continue
            if stream.get_stream_model() == 'mss':
                # 这里的init实际上是不正确的 这里生成是为了满足下载文件检查等逻辑
                stream.fix_header(is_fake=True)
            logger.info(f'{stream.get_name()} {t_msg.download_start}.')
            speed_up_flag = self.args.speed_up
            while max_failed > 0:
                loop = new_event_loop()
                results = loop.run_until_complete(self.do_with_progress(loop, stream, speed_up_flag))
                speed_up_flag = False
                loop.close()
                all_results.append(results)
                count_none, count_true, count_false = 0, 0, 0
                for _, flag in results.items():
                    if flag is True:
                        count_true += 1
                    elif flag is False:
                        count_false += 1
                    else:
                        count_none += 1
                # 出现False则说明无法下载
                if count_false > 0:
                    break
                # False 0 出现None则说明需要继续下载 否则合并
                if count_none > 0:
                    max_failed -= 1
                    continue
                break
            # track_id 最佳获取方案是从实际分段中提取 通过ism元数据无法直接计算出来
            if stream.get_stream_model() == 'mss':
                stream.fix_header(is_fake=False)
            self.try_concat(stream)
            # 只需要检查一个流的时间达到最大值就停止录制
            # 应当进行优化 只针对单个流进行停止录制
            if self.args.live and should_stop_record is False and stream.check_record_time(self.args.live_duration):
                should_stop_record = True
                logger.debug(f'set should_stop_record flag as {should_stop_record}')
        # 主动停止录制
        if should_stop_record:
            self.stop_record()
        return all_results

    def try_concat(self, stream: Stream):
        if self.args.live is False and self.args.disable_auto_concat is False:
            stream.concat(self.args)

    def try_concat_streams(self, streams: List[Stream], selected: List[str]):
        for stream in streams:
            if stream.get_skey() not in selected:
                continue
            if self.args.live is True and self.args.disable_auto_concat is False:
                stream.concat(self.args)

    def init_progress(self, stream: Stream, count: int, completed: int, speed_up_flag: bool):
        if completed > 0:
            if stream.filesize > 0:
                total = stream.filesize
            else:
                total = completed
                stream.filesize = total
        else:
            if stream.filesize > 0:
                total = stream.filesize
            else:
                total = 0
                stream.filesize = total
            completed = 0
        self.xprogress = XProgress(
            stream.get_name(),
            len(stream.segments),
            count,
            total,
            completed,
            speed_up_flag,
            self.args.speed_up_left,
        )

    async def do_with_progress(self, loop: AbstractEventLoop, stream: Stream, speed_up_flag: bool):
        '''
        下载过程输出进度 并合理处理异常
        '''
        results = {} # type: Dict[bool]
        tasks = set() # type: Set[Task]
        is_error = False

        def _done_callback(_future: Future) -> None:
            nonlocal results
            if _future.exception() is None:
                segment, status, flag = _future.result()
                if flag is None:
                    segment.content = []
                    # logger.error('下载过程中出现已知异常 需重新下载\n')
                elif flag is False:
                    segment.content = []
                    # 某几类已知异常 如状态码不对 返回头没有文件大小 视为无法下载 主动退出
                    cancel_all_task()
                    if status in ['STATUS_CODE_ERROR', 'NO_CONTENT_LENGTH']:
                        logger.error(f'{status} {t_msg.segment_cannot_download}')
                    elif status == 'EXIT':
                        pass
                    else:
                        logger.error(f'{status} {t_msg.segment_cannot_download_unknown_status}')
                results[segment] = flag
                if self.xprogress.is_ending():
                    cancel_uncompleted_task()
            else:
                # 出现未知异常 强制退出全部task
                logger.error(f'{t_msg.segment_cannot_download_unknown_exc} => {_future.exception()}\n')
                cancel_all_task()
                results['未知segment'] = False

        def cancel_uncompleted_task() -> None:
            cancel_all_task()
            _, _, _left = get_left_segments(stream)
            for segment in _left:
                if segment.max_retry_404 <= 0:
                    continue
                results[segment] = None
                segment.content = []

        def cancel_all_task() -> None:
            nonlocal is_error
            is_error = True
            for task in tasks:
                task.remove_done_callback(_done_callback)
            for task in filter(lambda task: not task.done(), tasks):
                task.cancel()
        # limit_per_host 根据不同网站和网络状况调整 如果与目标地址连接性较好 那么设置小一点比较好
        count, completed, _left = get_left_segments(stream)
        logger.debug(f'downloaded count {count}, downloaded size {completed}, left count {len(_left)}')
        if len(_left) == 0:
            return results
        # 剩余数量小于预期 不加速
        # 场景 => 在反复下载后还是少几个分段 然后重新跑命令下载
        # 这个时候如果开了加速 那么就会在刚开始下载的时候就重下
        if len(_left) <= self.args.speed_up_left:
            speed_up_flag = False
        # 没有需要下载的则尝试合并 返回False则说明需要继续下载完整
        self.init_progress(stream, count, completed, speed_up_flag)
        ts = time.time()
        client = ClientSession(connector=get_connector(self.args), timeout=ClientTimeout(total=None, sock_connect=15, sock_read=15)) # type: ClientSession
        for count, segment in enumerate(_left):
            if segment.max_retry_404 <= 0:
                self.xprogress.decrease_total_count()
                continue
            task = loop.create_task(self.download(client, stream, segment))
            task.add_done_callback(_done_callback)
            tasks.add(task)
            if self.args.gen_init_only and count == 0:
                break
        logger.info(f'{len(tasks)} tasks start')
        # 阻塞并等待运行完成
        finished, unfinished = await asyncio.wait(tasks)
        # 关闭ClientSession
        await client.close()
        self.xprogress.to_stop(is_error=is_error)
        logger.info(f'tasks end, time used {time.time() - ts:.2f}s')
        return results

    async def download(self, client: ClientSession, stream: Stream, segment: Segment):
        status, flag = 'EXIT', True
        try:
            async with client.get(segment.url + self.args.url_patch, headers=self.args.headers) as resp: # type: ClientResponse
                _flag = True
                if self.log_detail or resp.status != 200:
                    logger.debug(f'{segment.name} status {resp.status}, {segment.url + self.args.url_patch}')
                if resp.status in [403, 404]:
                    status = 'STATUS_SKIP'
                    flag = False
                    self.xprogress.decrease_total_count()
                    segment.skip_concat = True
                    if resp.status == 404:
                        segment.max_retry_404 -= 1
                if resp.status == 405:
                    status = 'STATUS_CODE_ERROR'
                    flag = False
                if resp.status in self.args.redl_code:
                    status = 'RE-DOWNLOAD'
                    flag = None
                if resp.headers.get('Content-length') is not None:
                    # 对于 filesize 不为 0 后面再另外考虑
                    size = int(resp.headers["Content-length"])
                    stream.filesize += size
                    logger.debug(f'{segment.name} response Content-length => {size}')
                    self.xprogress.update_total_size(stream.filesize)
                else:
                    size = -1
                    logger.debug(f'{segment.name} response header has no Content-length {dict(resp.headers)}')
                    _flag = False
                if flag:
                    while self.terminate is False:
                        data = await resp.content.read(512)
                        if not data:
                            break
                        segment.content.append(data)
                        self.xprogress.add_downloaded_size(len(data))
                        if _flag is False:
                            stream.filesize += len(data)
                            logger.debug(f'{segment.name} recv {size} byte data')
                            self.xprogress.update_total_size(stream.filesize)
        except TimeoutError:
            return segment, 'TimeoutError', None
        except client_exceptions.ClientConnectorError:
            return segment, 'ClientConnectorError', None
        except client_exceptions.ClientPayloadError:
            return segment, 'ClientPayloadError', None
        except client_exceptions.ClientOSError:
            return segment, 'ClientOSError', None
        except client_exceptions.ServerDisconnectedError:
            return segment, 'ServerDisconnectedError', None
        except client_exceptions.InvalidURL:
            return segment, 'EXIT', False
        except CancelledError:
            return segment, 'EXIT', False
        except Exception as e:
            logger.error(f'! -> {segment.url}', exc_info=e)
            return segment, status, False
        if self.terminate:
            return segment, 'EXIT', False
        if segment.skip_concat:
            return segment, status, True
        if flag is None:
            return segment, status, None
        if flag is False:
            return segment, status, False
        self.xprogress.add_downloaded_count(1)
        logger.debug(f'{segment.name} download end, size => {sum([len(data) for data in segment.content])}')
        return segment, 'SUCCESS', await self.decrypt(segment, stream)

    async def decrypt(self, segment: Segment, stream: Stream) -> bool:
        '''
        解密部分
        '''
        if self.args.disable_auto_decrypt is True:
            logger.debug(f'--disable-auto-decrypt, skip decrypt')
            return segment.dump()
        if segment.is_encrypt() and segment.is_supported_encryption():
            logger.debug(f'common aes decrypt, key {segment.xkey.key.hex()} iv {segment.xkey.iv}')
            cipher = CommonAES(segment.xkey.key, binascii.a2b_hex(segment.xkey.iv))
            return cipher.decrypt(segment)
        else:
            return segment.dump()



from typing import List
from hlsdl.cmdargs import CmdArgs
from hlsdl.extractors.base import BaseParser
from hlsdl.extractors.hls.ext.xkey import XKey
from hlsdl.extractors.hls.stream import HLSStream
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class HLSParser(BaseParser):
    def __init__(self, args: CmdArgs, uri_type: str):
        super(HLSParser, self).__init__(args, uri_type)
        self.suffix = '.m3u8'

    def parse(self, uri: str, content: str, parent_stream: HLSStream) -> List[HLSStream]:
        uri_item = self.parse_uri(uri)
        if uri_item is None:
            logger.error(f'parse {uri} failed')
            return []
        self.dump_content(uri_item.name, content, self.suffix)
        streams = []
        sindex = 0
        stream = HLSStream(sindex, uri_item, self.args.save_dir, parent_stream)
        stream.set_origin_url(uri_item.home_url, uri_item.base_url, uri)
        lines = [line.strip() for line in content.split('\n')]
        offset = 0
        last_segment_xkey = None # type: XKey
        content_is_master_type = False
        last_segment_has_xkey = False
        do_not_append_at_end_list_tag = False
        while offset < len(lines):
            segment = stream.segments[-1]
            line = lines[offset]
            # 分段标准tag参考 -> https://tools.ietf.org/html/rfc8216#section-4.3.2
            if line == '':
                pass
            elif line.startswith('#EXTM3U'):
                stream.set_tag('#EXTM3U')
            elif line.startswith('#EXT-X-VERSION'):
                pass
            elif line.startswith('#EXT-X-KEY'):
                if offset > 0 and lines[offset - 1].startswith('#EXT-X-'):
                    # 把这个位置的#EXT-X-KEY认为是全局的
                    stream.set_key(uri_item.home_url, uri_item.base_url, line)
                else:
                    segment.set_key(uri_item.home_url, uri_item.base_url, line)
                    if last_segment_has_xkey is False:
                        last_segment_has_xkey = True
                        last_segment_xkey = segment.get_xkey()
            elif line.startswith('#EXT-X-SESSION-KEY'):
                # 某些网站的
                pass
            elif line.startswith('#EXT-X-I-FRAMES-ONLY'):
                pass
            elif line.startswith('#EXT-X-INDEPENDENT-SEGMENTS'):
                pass
            elif line.startswith('#EXT-X-ALLOW-CACHE'):
                pass
            elif line.startswith('#EXT-X-MEDIA-SEQUENCE'):
                pass
            elif line.startswith('#EXT-X-PROGRAM-DATE-TIME'):
                # 根据spec 该标签指明的是第一个分段绝对日期/时间
                stream.set_xprogram_date_time(line)
            elif line.startswith('#EXT-X-DATERANGE'):
                # 按设计应该把这个标签认为是一个Stream的属性
                stream.set_daterange(line)
            elif line.startswith('#EXT-X-TARGETDURATION'):
                pass
            elif line.startswith('#EXT-X-PLAYLIST-TYPE'):
                pass
            elif line.startswith('#EXT-X-DISCONTINUITY'):
                if self.args.dont_split_discontinuity:
                    pass
                else:
                    # 此标签后面的分段都认为是一个新的Stream 直到结束或下一个相同标签出现
                    # 对于优酷 根据特征字符匹配 移除不需要的Stream 然后将剩余的Stream合并
                    sindex += 1
                    _xkey = stream.xkey
                    _bakcup_xkey = stream.bakcup_xkey
                    streams.append(stream)
                    stream = HLSStream(sindex, uri_item, self.args.save_dir, parent_stream)
                    stream.set_origin_url(uri_item.home_url, uri_item.base_url, uri)
                    stream.set_xkey(_xkey)
                    stream.set_bakcup_xkey(_bakcup_xkey)
                    stream.set_tag('#EXT-X-DISCONTINUITY')
            elif line.startswith('#EXT-X-MAP'):
                segment.set_map_url(uri_item.home_url, uri_item.base_url, line)
                stream.set_map_flag()
                stream.append_segment()
            elif line.startswith('#EXT-X-TIMESTAMP-MAP'):
                pass
            elif line.startswith('#USP-X-TIMESTAMP-MAP'):
                pass
            elif line.startswith('#EXTINF'):
                segment.set_duration(line)
            elif line.startswith('#EXT-X-PRIVINF'):
                segment.set_privinf(line)
            elif line.startswith('#EXT-X-BYTERANGE'):
                segment.set_byterange(line)
            elif line.startswith('#EXT-X-BITRATE'):
                pass
            elif line.startswith('#EXT-X-ENDLIST'):
                pass
            elif line.startswith('#EXT-X-MEDIA'):
                # 外挂媒体 视为单独的一条流
                sindex += 1
                stream.set_tag('#EXT-X-MEDIA')
                stream.set_media(uri_item.home_url, uri_item.base_url, line)
                content_is_master_type = True
                streams.append(stream)
                stream = HLSStream(sindex, uri_item, self.args.save_dir, parent_stream)
            # elif line.startswith('#EXT-X-STREAM-INF'):
            elif line.startswith('#EXT-X-') and 'STREAM-INF' in line:
                stream.set_tag('#EXT-X-STREAM-INF')
                stream.set_xstream_inf(line)
                content_is_master_type = True
                if stream.xstream_inf.uri is None:
                    pass
                else:
                    # handle for #EXT-X-I-FRAME-STREAM-INF
                    sindex += 1
                    do_not_append_at_end_list_tag = True
                    stream.set_origin_url(uri_item.home_url, uri_item.base_url, stream.xstream_inf.uri)
                    streams.append(stream)
                    stream = HLSStream(sindex, uri_item, self.args.save_dir, parent_stream)
            elif line.startswith('#'):
                if line.startswith('## Generated with https://github.com/google/shaka-packager'):
                    pass
                elif line.startswith('## Created with Unified Streaming Platform'):
                    pass
                else:
                    logger.warning(f'unknown TAG, skip\n\t{line}')
            else:
                # 进入此处 说明这一行没有任何已知的#EXT标签 也就是具体媒体文件的链接
                if offset > 0 and lines[offset - 1].startswith('#EXT-X-BYTERANGE'):
                    segment.set_xkey(last_segment_has_xkey, last_segment_xkey)
                    segment.set_url(uri_item.home_url, uri_item.base_url, line)
                    stream.append_segment()
                elif offset > 0 and lines[offset - 1].startswith('#EXT-X-PRIVINF'):
                    segment.set_xkey(last_segment_has_xkey, last_segment_xkey)
                    segment.set_url(uri_item.home_url, uri_item.base_url, line)
                    stream.append_segment()
                elif offset > 0 and lines[offset - 1].startswith('#EXTINF') or lines[offset - 1].startswith('#EXT-X-BITRATE'):
                    segment.set_xkey(last_segment_has_xkey, last_segment_xkey)
                    segment.set_url(uri_item.home_url, uri_item.base_url, line)
                    stream.append_segment()
                elif offset > 0 and lines[offset - 1].startswith('#EXT-X-') and 'STREAM-INF' in lines[offset - 1]:
                    sindex += 1
                    stream.set_url(uri_item.home_url, uri_item.base_url, line)
                    streams.append(stream)
                    stream = HLSStream(sindex, uri_item, self.args.save_dir, parent_stream)
                    do_not_append_at_end_list_tag = True
                else:
                    logger.warning(f'unknow what to do here ->\n\t{line}')
            offset += 1
        if do_not_append_at_end_list_tag is False:
            streams.append(stream)
        # 下面的for循环中stream/segment是浅拷贝
        _stream_paths = []
        _streams = [] # type: List[HLSStream]
        for stream in streams:
            # 去重
            if stream.tag == '#EXT-X-STREAM-INF':
                stream_path = stream.get_path()
                if stream_path in _stream_paths:
                    continue
                elif stream_path == '':
                    # 文件类型
                    pass
                else:
                    _stream_paths.append(stream_path)
            # 处理掉末尾空的分段
            if stream.segments[-1].url == '':
                _ = stream.segments.pop(-1)
            # 过滤掉广告片段
            _segments = []
            for segment in stream.segments:
                if self.args.ad_keyword == '':
                    _segments.append(segment)
                elif self.args.ad_keyword not in segment.url:
                    _segments.append(segment)
            stream.segments = _segments
            # 保留过滤掉广告片段分段数大于0的Stream
            if len(stream.segments) > 0 or stream.tag == '#EXT-X-STREAM-INF' or stream.tag == '#EXT-X-MEDIA':
                _streams.append(stream)
        # if content_is_master_type is False and len(_streams) > 1:
        if content_is_master_type is False and len(_streams) > 1:
            streams = []
            # 合并去除#EXT-X-DISCONTINUITY后剩下的Stream
            stream = _streams[0]
            for _stream in _streams[1:]:
                # 一条流中有map时不参与合并
                if _stream.has_map_segment:
                    streams.append(stream)
                    stream = _stream
                    continue
                else:
                    stream.segments_extend(_stream.segments, name_from_url=self.args.name_from_url)
            streams.append(stream)
            _streams = streams
        return _streams


import time
from typing import List
from pathlib import Path
from hlsdl.cmdargs import CmdArgs
from hlsdl.extractor import Extractor
from hlsdl.downloader import Downloader
from hlsdl.models.stream import Stream
from hlsdl.extractors.hls.stream import HLSStream
from hlsdl.extractors.dash.stream import DASHStream
from hlsdl.extractors.dash.parser import DASHParser
from hlsdl.extractors.dash.childs.location import Location
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class Daemon:

    def __init__(self, args: CmdArgs):
        self.args = args
        self.exit = False

    def daemon(self):
        '''
        - 解析
        - 下载
        - 合并
        '''
        extractor = Extractor(self.args)
        streams = extractor.fetch_metadata(self.args.URI[0])
        if self.args.live is False:
            return Downloader(self.args).download_streams(streams)
        else:
            return self.live_record(extractor, streams)

    def live_record(self, extractor: Extractor, streams: List[Stream]):
        '''
        - 第一轮解析 判断类型
        - 选择流 交给下一个函数具体下载-刷新-下载...
        - TODO
        -   提供录制时长设定
        '''
        if isinstance(streams[0], DASHStream):
            self.live_record_dash(extractor, streams)
        if isinstance(streams[0], HLSStream):
            self.live_record_hls(extractor, streams)
        # assert False, f'unsupported live stream type {type(streams[0])}'

    def live_record_dash(self, extractor: Extractor, streams: List[DASHStream]):
        '''
        dash直播流
        重复拉取新的mpd后
        判断是否和之前的重复关键在于url的path部分是不是一样的
        也就是说 文件是不是一个
        那么主要逻辑如下
        - 再次解析 这一轮解析的时候 select 复用第一轮的选择 用skey来判断
        - 判断合并 下载已经解析好的分段
        - 再次解析 再次下载 直到满足结束条件
        Q 为何先再次解析而不是先下载完第一轮解析的分段再下载
        A 第一轮解析时有手动选择流的过程 而dash流刷新时间一般都很短 往往只有几秒钟 所以最好是尽快拉取一次最新的mpd
        Q 为什么不拉取新mpd单独开一个线程
        A 下载的时候很可能会占满网速 个人认为循环会好一点
        Q 万一下载卡住导致mpd刷新不及时怎么办
        A 还没有想好 不过这种情况概率蛮小的吧... 真的发生了说明你的当前网络不适合录制
        '''
        # assert False, 'not support dash live stream, wait plz'
        # 再次解析 优先使用 Location 作为要刷新的目标链接
        # 因为有的直播流 Location 会比用户填写的链接多一些具体标识 比如时间 或者token
        parser = extractor.parser # type: DASHParser
        locations = parser.root.find('Location') # type: List[Location]
        if len(locations) == 1:
            next_mpd_url: str = locations[0].innertext.strip()
        else:
            next_mpd_url = self.args.URI[0]
        if '://' not in next_mpd_url:
            if Path(next_mpd_url).is_file() or Path(next_mpd_url).is_dir():
                assert False, 'not support dash live stream for file/folder type, because cannot refresh'
        logger.info(f'refresh link {next_mpd_url}')
        # 初始化下载器
        downloader = Downloader(self.args)
        # 获取用户选择的流的skey
        skeys = downloader.do_select(streams)
        if len(skeys) == 0:
            return
        refresh_interval = self.args.live_refresh_interval
        last_time = time.time()
        while True:
            # 刷新间隔时间检查
            if time.time() - last_time < refresh_interval:
                time.sleep(0.5)
                continue
            last_time = time.time()
            # 复用 extractor 再次解析
            # 这里不应该是文件或者文件夹 当然第一轮可以是链接和文件
            next_streams = extractor.fetch_metadata(next_mpd_url)
            # 合并下载分段信息
            self.streams_extend(streams, next_streams, skeys)
            # 下载分段
            downloader.download_streams(streams, selected=skeys)
            # 检查是不是主动退出了
            if downloader.terminate:
                logger.debug(f'downloader terminated break')
                break
            # 继续循环
        downloader.try_concat_streams(streams, skeys)

    def live_record_hls(self, extractor: Extractor, streams: List[HLSStream]):
        '''
        hls直播流
        '''
        assert False, 'not support hls live stream, wait plz'

    def streams_extend(self, streams: List[DASHStream], next_streams: List[DASHStream], skeys: List[str]):
        _streams = dict((stream.get_skey(), stream) for stream in streams)
        _next_streams = dict((stream.get_skey(), stream) for stream in next_streams)
        for skey in skeys:
            _stream = _streams.get(skey) # type: DASHStream
            _next_stream = _next_streams.get(skey) # type: DASHStream
            if _stream is None or _next_stream is None:
                continue
            # 对于新增的分段 认为默认有init分段
            _stream.live_segments_extend(_next_stream.segments, has_init=True, name_from_url=self.args.name_from_url, compare_with_url=self.args.compare_with_url)


from ..ismitem import ISMItem


class c(ISMItem):
    def __init__(self, name: str):
        super(c, self).__init__(name)
        self.t = None # type: int
        self.d = None # type: int
        self.r = 1 # type: int

    def generate(self):
        if self.t is not None:
            self.to_int('t')
        self.to_int('d')
        self.to_int('r')


import os
import platform
from typing import List
from pathlib import Path
from hlsdl.cmdargs import CmdArgs
from hlsdl.util.texts import t_msg

ONCE_MAX_FILES = 500


class Concat:

    @staticmethod
    def call_mp4decrypt(out_path: Path, args: CmdArgs):
        assert out_path.exists() is True, f'File not exists ! -> {out_path}'
        assert out_path.stat().st_size > 0, f'File concat failed ! -> {out_path}'
        out = out_path.absolute().as_posix()
        out_decrypted = (out_path.parent / f'{out_path.stem}_decrypted{out_path.suffix}').absolute()
        if args.overwrite is False and out_decrypted.exists():
            print(t_msg.decrypted_file_exists_skip)
            return
        if platform.system() == 'Windows':
            _cmd = f'""{args.mp4decrypt}" --show-progress --key {args.key} "{out}" "{out_decrypted.as_posix()}""'
        else:
            _cmd = f'"{args.mp4decrypt}" --show-progress --key {args.key} "{out}" "{out_decrypted.as_posix()}"'
        print(t_msg.start_decrypt)
        os.system(_cmd)
        if args.enable_auto_delete:
            if out_decrypted.exists() and out_decrypted.stat().st_size > 0:
                os.remove(out)

    @staticmethod
    def gen_new_names(names: list, out: str, tmp_suffix: str = '.tmp'):
        work_num = len(names) // ONCE_MAX_FILES + 1
        counts = len(names) // work_num
        new_names = []
        _tmp_outs = []
        for multi_index in range(work_num):
            if multi_index < work_num - 1:
                _names = names[multi_index * counts:(multi_index + 1) * counts]
            else:
                _names = names[multi_index * counts:]
            _tmp_outs.append(f'out{multi_index}{tmp_suffix}')
            new_names.append([_names, f'out{multi_index}{tmp_suffix}'])
        new_names.append([_tmp_outs, out])
        return new_names, _tmp_outs

    @staticmethod
    def gen_cmds_outs(out_path: Path, names: list, args: CmdArgs) -> List[str]:
        out = out_path.absolute().as_posix()
        cmds = [] # type: List[str]
        if args.raw_concat is False:
            if len(names) > ONCE_MAX_FILES:
                new_names, _tmp_outs = Concat.gen_new_names(names, out, tmp_suffix=".ts")
                for _names, _out in new_names:
                    if platform.system() == 'Windows':
                        cmds.append(f'""{args.ffmpeg}" -i concat:"{"|".join(_names)}" -c copy -y "{_out}" > nul"')
                    else:
                        cmds.append(f'"{args.ffmpeg}" -i concat:"{"|".join(_names)}" -c copy -y "{_out}" > nul')
                return cmds, _tmp_outs
            if platform.system() == 'Windows':
                return [f'""{args.ffmpeg}" -i concat:"{"|".join(names)}" -c copy -y "{out}" > nul"'], []
            else:
                return [f'"{args.ffmpeg}" -i concat:"{"|".join(names)}" -c copy -y "{out}" > nul'], []
        if len(names) > ONCE_MAX_FILES:
            new_names, _tmp_outs = Concat.gen_new_names(names, out)
            if platform.system() == 'Windows':
                for _names, _out in new_names:
                    cmds.append(f'copy /b {"+".join(_names)} "{_out}" > nul')
                return cmds, _tmp_outs
            else:
                for _names, _out in new_names:
                    cmds.append(f'cat {" ".join(_names)} > "{_out}"')
                return cmds, _tmp_outs
        if platform.system() == 'Windows':
            return [f'copy /b {"+".join(names)} "{out}"'], []
        else:
            return [f'cat {" ".join(names)} > "{out}"'], []


from xml.parsers.expat import ParserCreate
from .ism import ISM
from .childs.c import c
from .childs.protection import Protection
from .childs.protectionheader import ProtectionHeader
from .childs.qualitylevel import QualityLevel
from .childs.streamindex import StreamIndex


def xml_handler(content: str):
    def handle_start_element(tag, attrs):
        nonlocal ism
        nonlocal ism_handlers
        if ism is None:
            if tag != 'SmoothStreamingMedia':
                raise Exception('the first tag is not SmoothStreamingMedia!')
            ism = ISM(tag)
            ism.addattrs(attrs)
            ism.generate()
            stack.append(ism)
        else:
            if ism_handlers.get(tag) is None:
                return
            child = ism_handlers[tag](tag)
            child.addattrs(attrs)
            if tag != 'ProtectionHeader':
                child.generate()
            ism.childs.append(child)
            ism = child
            stack.append(child)

    def handle_end_element(tag):
        nonlocal ism
        nonlocal ism_handlers
        if ism_handlers.get(tag) is None:
            return
        if tag == 'ProtectionHeader':
            ism.generate()
        if len(stack) > 1:
            _ = stack.pop(-1)
            ism = stack[-1]

    def handle_character_data(texts: str):
        if texts.strip() != '':
            ism.innertext += texts.strip()
    stack = []
    ism = None # type: ISM
    ism_handlers = {
        'SmoothStreamingMedia': ISM,
        'StreamIndex': StreamIndex,
        'QualityLevel': QualityLevel,
        'c': c,
        'Protection': Protection,
        'ProtectionHeader': ProtectionHeader,
    }
    parser = ParserCreate()
    parser.StartElementHandler = handle_start_element
    parser.EndElementHandler = handle_end_element
    parser.CharacterDataHandler = handle_character_data
    parser.Parse(content)
    return ism


from ..mpditem import MPDItem


class SegmentBaee(MPDItem):

    def __init__(self, name: str):
        super(SegmentBaee, self).__init__(name)
        self.indexRange = ""
        self.timescale = None # type: int
        self.presentationTimeOffset = None # type: int

    def generate(self):
        if self.presentationTimeOffset is None:
            self.presentationTimeOffset = 0
        if self.timescale is None:
            self.timescale = 0
        self.to_int()

    def to_int(self):
        self.timescale = int(self.timescale)
        self.presentationTimeOffset = int(self.presentationTimeOffset)


from datetime import datetime
from .x import X


class XDateRange(X):
    '''
    #EXT-X-DATERANGE 第一个分段的绝对时间
    - 2019-01-01T00:00:00.000Z
    '''
    def __init__(self):
        super(XDateRange, self).__init__('#EXT-X-DATERANGE')
        self._id = None # type: str
        self._class = None # type: str
        self.start_date = None # type: datetime
        self.end_date = None # type: datetime
        self.duration = None # type: float
        self.planned_duration = None # type: float
        self.end_on_next = None # type: str
        self.known_attrs = {
            'ID': '_id',
            'CLASS': '_class',
            'START-DATE': self.set_start_date,
            'END-DATE': self.set_end_date,
            'DURATION': self.set_duration,
            'PLANNED-DURATION': self.set_planned_duration,
            'END-ON-NEXT': 'end_on_next',
        }

    def set_duration(self, text: str):
        self.duration = float(text)

    def set_planned_duration(self, text: str):
        self.planned_duration = float(text)

    def get_time(self, text: str):
        if text.endswith('Z') is True:
            text = f'{text[:-1]}+00:00'
        try:
            time = datetime.fromisoformat(text)
        except Exception:
            raise
        return time

    def set_start_date(self, text: str):
        self.start_date = self.get_time(text)

    def set_end_date(self, text: str):
        self.end_date = self.get_time(text)

    def set_attrs_from_line(self, line: str):
        # https://stackoverflow.com/questions/127803
        # datetime.strptime(line, "%Y-%m-%dT%H:%M:%S.%fZ")
        info = self.get_tag_info(line)
        for key, value in self.regex_attrs(info):
            value = value.strip('"')
            if key in self.known_attrs:
                if isinstance(self.known_attrs[key], str):
                    self.__setattr__(self.known_attrs[key], value)
                elif isinstance(self.known_attrs[key], type):
                    self.convert_type(key, value, self.known_attrs[key])
                else:
                    self.known_attrs[key](value)
            elif key.startswith('X-'):
                self.__setattr__(self.format_key(key), value)
            else:
                print(f'unknown attr of {self.TAG_NAME}')
        return self